package mil.navy.nps.bridge;

import java.io.*;
import java.net.*;


/*
   BridgeMulticastToUnicast

   This is a sublcass of the Bridge abstract class, which implements datagram
   bridging between a multicast socket and a unicast socket. The program simply 
   accepts packets on a multicast socket, and repeats them out on a unicast
   socket.

    HISTORY

	20Aug97	DMcG			Made "multicast" terminology consistent


   @author Don McGregor (<A HREF="http://www.stl.nps.navy.mil/~mcgredo"> http://www.stl.nps.navy.mil/~mcgredo</A>)<P>
   @version 1.0 
   
*/

public class BridgeMulticastToUnicast extends Bridge implements Runnable
{

  private MulticastSocket inputSocket             = null;   // socket we listen on 
  private DatagramSocket  outputSocket            = null;   // socket we write to

  private int             inputSocketNumber       = -1;     // input socket number 
  private String          multicastAddress        = null;   // multicast Address
  private int             outputSocketNumber      = -1;     // output socket number


  
 public BridgeMulticastToUnicast( 
        int     pInputSocketNumber,                         // INPUT: socket to listen on
        String  pMulticastAddress,                          // INPUT: multicast Address to join
        int     pOutputSocketNumber,                        // INPUT: socket we write on
        String  pDestMachineName,                           // INPUT: name of machine we're sending to
        int     pDestSocketNumber)                          // INPUT: socket to write to on foreign machine
                        
{

   // superclass initialization

   super(pDestMachineName, pDestSocketNumber);

  multicastAddress = pMulticastAddress;

  System.out.println ("BridgeMulticastToUnicast multicastAddress   = " + pMulticastAddress);
  System.out.println ("BridgeMulticastToUnicast multicastPort      = " + pInputSocketNumber);
  System.out.println ("BridgeMulticastToUnicast   output port      = " + pOutputSocketNumber);
  System.out.println ("BridgeMulticastToUnicast destinationAddress = " + pDestMachineName);
  System.out.println ("BridgeMulticastToUnicast destinationPort    = " + pDestSocketNumber);

  try
    {
        inputSocket  = new MulticastSocket(pInputSocketNumber);
        inputSocket.joinGroup(InetAddress.getByName(pMulticastAddress));
    }

  // unkHostException has to be caught first, since it's more specific than an
  // IOException.

  catch(UnknownHostException unkHostException)
    {
      throw new RuntimeException("problem with joining input multicast Address " + multicastAddress);
    }
  catch(SocketException sockEx)
    {
      throw new RuntimeException("couldn't create inputSocket");
    }
  catch(IOException ioEx)
    {
      throw new RuntimeException("problem with input multicast socket");
    }
  
  try
    {
        outputSocket = new DatagramSocket(pOutputSocketNumber);
    }

  // unkHostException has to be caught first, since it's more specific than an
  // IOException.

  catch(SocketException sockEx)
    {
      throw new RuntimeException("couldn't create outputSocket");
    }
  catch(IOException ioEx)
    {
      throw new RuntimeException("problem with output unicast socket");
    }
  
  
   return;
 }

 public void run()
 {
   // listen on a multicast socket, repeat it on a unicast socket.

  DatagramPacket  inDgram   = null;
  DatagramPacket  outDgram  = null;
  byte            ibuff[]   = null;
  byte            obuff[]   = null;

  while(true)
  {
    ibuff   = new byte[MAX_PACKET_SIZE];
    obuff   = new byte[MAX_PACKET_SIZE];
    inDgram = new DatagramPacket(ibuff, ibuff.length);

    if(DEBUG)
      System.out.println("reading...");

    // Read a datagram into the new buffer.
    try
    {
      inputSocket.receive(inDgram);
    }
    catch(IOException ioe)
    {
      throw new RuntimeException("error reading from socket");
    }

    if(DEBUG)
      System.out.println("writing...");

    // create a new outgoing datagram from the incoming datagram's data. Note that
    // we have to be careful about length here; if they don't match up on the sending
    // and receiving side, we'll run into problems. The "solution" for now is to always
    // use datagrams of length MAX_PACKET_SIZE. Need to find a better way around this,
    // but don't have time to look at it right now.

    System.arraycopy(ibuff, 0, obuff, 0, ibuff.length);

    outDgram = new DatagramPacket(obuff,                      // data from incoming datagram
                                  obuff.length,               // length of data
                                  destMachineInet,            // machine we're sending to
                                  destinationSocketNumber);   // socket on machine we're sending to

    if(DEBUG)
    {
      System.out.println("outDgram length = " + outDgram.getLength());
      System.out.println("outDgram destination socket number = " + destinationSocketNumber);
    }

    try
    {
      outputSocket.send(outDgram);
    }
    catch(IOException ioe)
    {
      throw new RuntimeException("error sending to output socket");
    }

  } // end of endless loop

} // end of run()

} // end of class
