/*
 File:		AcknowledgePdu.java
 CVS Info:	$Id: AcknowledgePdu.java,v 1.2 1998/01/27 18:43:49 mcgredo Exp $
 Compiler:	jdk1.1.6
 */

package mil.navy.nps.dis;                                       // package to which we belong

import mil.navy.nps.util.*;         // General-purpose utilities
import mil.navy.nps.disEnumerations.*;         // Enumerations for DIS

import java.util.*;                                             // utility stuff we need
import java.io.*;                                               // input/output for serialization

import iwt.*;
import org.w3c.dom.*;
import com.docuverse.freedom.*;

/**
 *@version 1.0
 *@author Antonio Alexandre Rua (<A HREF="http://www.garfield.fe.up.pt/~alexrua">http://www.garfield.fe.up.pt/~alexrua</A>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/AcknowledgePdu.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/AcknowledgePdu.java</a>
 *
 *<dt><b>Hierarchy Diagram:</b>
 *<dd><a href="images/PduClassHierarchy.jpg"><IMG SRC="images/PduClassHierarchyButton.jpg" ALIGN=ABSCENTER></a>
 *
 *<dt><b>Summary:</b>
 *<dd>The acknowledgment of the receipt of a Start/Resume PDU, Stop/Freeze PDU, Create Entity PDU,
 *  or a Remove Entity PDU shall be communicated by issuing an Acknowledge PDU.
 *
 *<dt><b>Explanation:</b>
 *<dd>As the other Simulation Manager Pdu, this PDU has two header section, 
 *  the first one is inherited from the ProtocolDataUnit
 *  abstract class, the second from the SimulationManagementPDU abstract class.
 *  After That the PDU constains Acknowledge Flag, Response Flag and requestID fields.
 *  As with other things, the AcknowledgePdu has to know how to serialize and
 *  deserialize itself, clone itself, and print out its values.<P>
 *
 *<dt><b>History:</b>
 *<dd>		16Sep97	/Antonio Alexandre Rua	/New
 *<dd>		8Dec97	/Ronan Fauglas		/changes for documentation templates + complements in documentation
 *<dd>		11Dec97	/Ronan Fauglas		/changed access methods: thisVariable() --> getThisVariable()
 *<dd>		17Dec97	/Ronan Fauglas		/bug fix changed sizeOf 40->32
 *
 *<dt><b>References:</b>
 *<dd>		DIS Data Dictionary :<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/a3.htm">
 *		Acknowledge PDU</a>	
 *<dd>		DIS-Java-VRML Working Group: <a href="http://www.stl.nps.navy.mil/dis-java-vrml/"> 
 *		http://www.stl.nps.navy.mil/dis-java-vrml/</a>
 *<dd>		DIS specification : IEEE 1278.1, Section 5.4.6.5, 4.4.5.4.5
 *<dd>		See Section 7 in EBV-DOC
 *
 *@see ProtocolDataUnit
 *@see PduElement 
 *@see SerializationInterface
 *@see CreateEntityPdu
 *@see RemoveEntityPdu
 *@see StartResumePdu
 *@see StopFreezePdu
 *@see AcknowledgePdu
 *@see ActionRequestPdu
 *@see ActionResponsePdu
 *@see DataQueryPdu
 *@see SetDataPdu
 *@see DataPdu
 *@see EventReportPdu
 *@see CommentPdu
 */
public class AcknowledgePdu extends SimulationManagementPdu
{
    /**
     *Acknowledge Flag - This field shall indicate what type of message has been acknowledged.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>Enumeration, see references below for values.
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary: <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/a4.htm">
     *		Acknowledge Flag Field</a>
     *<dd>	See Section 7 in EBV-DOC
     *</dl>
     */
    protected UnsignedShort acknowledgeFlag;

    /**
     *Reponse flag - This field shall indicate whether or not the receiving entity was able 
     *to comply with the request, and for what reason.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>Enumeration, see references below for values.
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary: <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/a5.htm">
     *		Response Flag Field</a>
     *<dd>	See Section 7 in EBV-DOC
     *</dl>
     */
    protected UnsignedShort responseFlag;

    /**
     *Request ID - This field shall identify the matching response to the specific a Start/Resume,
     *Stop/Freeze, Create Entity, or Remove Entity PDU sent by the simulation manager.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>A 32-bit monotonically increasing integer identifier inserted by the Simulation Manager
     *  into all Simulation Manager PDUs. 
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary: <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/99.htm">
     *		Request ID Field</a>
     *</dl>
     */
    protected UnsignedInt requestID;        //identifies entity creation request by SM
      
    /**
     *Constant value--size of a full Acknowledge PDU with header.
     *<code>sizeOf = 32 bytes</code>
     */
    public static final int sizeOf = 32; // The total size of PDU is known

    /**
     *An "exemplar" object, which is filled out to the state that is needed most of the time.
     *
     *<dl>
     *<dd>
     *<dt><b>Explanation</b>
     *<dd>A brand new object has to have most of its values set,
     *  such as the forceID, protocol version, and so on. This lets the user fill
     *  out most of the values, save it in the class, then retrieve a copy of it
     *  as needed. 
     *</dd>
     *</dl>
     */
    static protected AcknowledgePdu exemplar;

/**
 *Default constructor--fills with zeros for all values.
 */
public AcknowledgePdu()                                             // default constructor
{
  setPduType(PDUTypeField.ACKNOWLEDGE);
  acknowledgeFlag = new UnsignedShort();
  responseFlag    = new UnsignedShort();
  requestID       = new UnsignedInt();

  return;
}

public void  serialize(DataOutputStream outputStream)
{
 /**
    Serialize our data out to the stream. We first call the superclass,
    so that all the data there is written out to the buffer first.
    Then we serialize each of our ivars to the stream, being careful
    about the order.
*/

  // takes care of all the header variables in the two consecutive superclasses
  super.serialize(outputStream);          
  // write out ivars
  acknowledgeFlag.serialize(outputStream);
  responseFlag.serialize(outputStream);
  requestID.serialize(outputStream);

  return;
}

public void  deSerialize(DataInputStream inputStream)
{
 /**
    deserialize our data from the stream. We first call the superclass,
    so that all the data there is read into the buffer first.
    Then we deserialize each of our ivars to the stream, being careful
    about the order.
 */

  super.deSerialize(inputStream);
  // Do our ivars
  acknowledgeFlag.deSerialize(inputStream);
  responseFlag.deSerialize(inputStream);
  requestID.deSerialize(inputStream);

  return;
}

public Object clone()
{
 /**
    Clone the CommentPdu, being careful to not share any pointers between the
    new object and the old object.
  */

  AcknowledgePdu newPdu = (AcknowledgePdu)super.clone();     // new Create Entity Pdu pdu

  newPdu.setAcknowledgeFlag(this.getAcknowledgeFlag());
  newPdu.setResponseFlag(this.getResponseFlag());
  newPdu.setRequestID(this.getRequestID());

  return newPdu;
}

public int length()
{
  return sizeOf;
}

public String PDUName()
{
  return new String("Acknowledge PDU");
}

public void printValues(int indentLevel, PrintStream printStream)
{
   // print the values of the object out, with correct level of
  // indentation on the page.
  char[]  indent = new char[indentLevel];
  int     idx = 0;
  int     superclassIndent = indentLevel;     // level of indentation for our superclass
  String  spacing;

  for(idx = 0; idx < indentLevel; idx++)
      indent[idx] = ' ';
  spacing = new String(indent);
  
  printStream.println();
  printStream.println("Acknowledge PDU");

  // ugly wart: get the superclass over to the left a couple pixels, if we have any to spare,
  // so the header info will be indented a bit less.
  if(superclassIndent > 0)
    superclassIndent -= 1;
  super.printValues(superclassIndent, printStream);
  
  printStream.println(indent + "Acknowledge Flag: "+acknowledgeFlag.intValue());
  printStream.println(indent + "Response Flag   : "+responseFlag.intValue());
  printStream.println(indent + "Request ID      : "+requestID.longValue());

  return;
}

public AcknowledgePdu getExemplar()
{
    return (AcknowledgePdu)exemplar.clone();
}
public void setExemplar(AcknowledgePdu newExemplar)
{
    exemplar = newExemplar;
}

  // Accessor methods
  public UnsignedShort getAcknowledgeFlag()
  {  return (UnsignedShort)acknowledgeFlag.clone(); }

  public void setAcknowledgeFlag(UnsignedShort pAcknowledgeFlag)
  {  acknowledgeFlag = pAcknowledgeFlag;}

  public void setAcknowledgeFlag(int pAcknowledgeFlag)
  {  acknowledgeFlag = new UnsignedShort(pAcknowledgeFlag);}

  public UnsignedShort getResponseFlag()
  {  return (UnsignedShort)responseFlag.clone();}

  public void setResponseFlag(UnsignedShort pResponseFlag)
  {  responseFlag = pResponseFlag;}

  public void setResponseFlag(int pResponseFlag)
  {  responseFlag = new UnsignedShort(pResponseFlag);}

  public UnsignedInt getRequestID()
  {  return (UnsignedInt)requestID.clone();}

  public void setRequestID(UnsignedInt pRequestID)
  {  requestID = pRequestID;}
  
  public void setRequestID(int pRequestID)
  {  requestID = new UnsignedInt(pRequestID); }

  /**
  */
  public void 
  write(org.w3c.dom.Document outDoc) 
  {  insert(outDoc);
  }// end of write

  /**
  */
  public void
  insert(org.w3c.dom.Node outNode) 
  {
    Element         ele;
    Element         ele1;
   
    Document outDoc = outNode.getOwnerDocument();
    ele1 = outDoc.createElement(iwt.DIS_TAGS.PDU);
    outNode.appendChild(ele1);

    // write out header info
    insertHeader(ele1);

    ele = outDoc.createElement(iwt.DIS_TAGS.RequestID);
    ele.setAttribute(iwt.DIS_TAGS.Value,requestID.toString());
    ele.setAttribute(iwt.DIS_TAGS.AcknowledgeFlag,acknowledgeFlag.toString());
    ele.setAttribute(iwt.DIS_TAGS.ResponseFlag,responseFlag.toString());
    ele1.appendChild(ele);

  }// end of insert

  /**
  */
  public void
  read(org.w3c.dom.Document inDoc) throws InvalidDocumentType 
  {
  }// end of insert

  /**
  */
  public void
  extract(org.w3c.dom.Node inNode)  throws InvalidDocumentType
  {
     Element         ele;
     NodeList        nodes;
     NodeList        aList;
     int i=0;

     if( !inNode.getNodeName().equals(iwt.DIS_TAGS.PDU) )
        throw new InvalidDocumentType();

     nodes = inNode.getChildNodes();

     while( i<nodes.getLength() )
     {
        Node aNode = nodes.item(i);

        if( aNode!=null )
        {
  
                if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                    aNode.getNodeName().equals(iwt.DIS_TAGS.DIS_PDU_Header) )
                    extractHeader(aNode);

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.RequestID) )
                 {
                   org.w3c.dom.NamedNodeMap attribs = aNode.getAttributes();
                   org.w3c.dom.Attribute a1 = (org.w3c.dom.Attribute) attribs.getNamedItem(iwt.DIS_TAGS.Value);
                   org.w3c.dom.Attribute a2 = (org.w3c.dom.Attribute) attribs.getNamedItem(iwt.DIS_TAGS.AcknowledgeFlag);
                   org.w3c.dom.Attribute a3 = (org.w3c.dom.Attribute) attribs.getNamedItem(iwt.DIS_TAGS.ResponseFlag);
       

                   String v1, v2, v3;

                   v1 = a1.getValue();
                   v2 = a2.getValue(); v3 = a3.getValue();

                   try
                   { requestID = new UnsignedInt(Integer.parseInt(v1));
                   } catch(NumberFormatException nfe) {}

                   try
                   { acknowledgeFlag  = new UnsignedShort(Integer.parseInt(v2));
                   } catch(NumberFormatException nfe) {}

                   try
                   { responseFlag = new UnsignedShort(Integer.parseInt(v3));
                   } catch(NumberFormatException nfe) {}

                 }
          
        }// end of if

        i++;

     }// end of while

  }// end of insert

}// end of class AcknowledgePdu

