/*
 File:		ArticulationParameter.java
 CVS Info:	$Id: ArticulationParameter.java,v 1.2 1998/01/27 18:43:54 mcgredo Exp $
 Compiler:	jdk 1.1.6 
 */

package mil.navy.nps.dis;                       // package we belong to

import mil.navy.nps.util.*;         // General-purpose utilities
import mil.navy.nps.disEnumerations.*;         // Enumerations for DIS

import java.io.*;                               // input/output library
import java.util.*;                             // utilities

import iwt.*;
import org.w3c.dom.*;

/** 
 *@version 1.0
 *@author <a href="mailto:mcgredo@stl.nps.navy.mil"> Don McGregor</a> (<a href="http://www.stl.nps.navy.mil/~mcgredo">http://www.stl.nps.navy.mil/~mcgredo</a>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/ArticulationParameter.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/ArticulationParameter.java</a>
 *
 *<dt><b>Hierarchy Diagram:</b>
 *<dd><a href="images/PduClassHierarchy.jpg"><IMG SRC="images/PduClassHierarchyButton.jpg" ALIGN=ABSCENTER></a>
 *
 *<dt><b>Summary:</b>
 *<dd>Location of the origin of the entity's coordinate system shall be specified 
 *  by a set of three coordinates: X, Y, and Z. 
 *
 *<dt><b>Explanation</b>
 *<dd>The articulation parameter describes how a part of an entity, such as a turret or fin,
 *  is oriented wrt the rest of the entity. <p>
 *
 *  This is a subclass of the PduElement class, the abstract class that all parts of a
 *  PDU inherit from. It knows how to serialize, deserialize, and clone itself. It
 *  is kept as a part of the EntityStatePdu.<P>
 *
 *<dt><b>History:</b>
 *<dd>		21Oct96 /Don McGregor    	/New
 *<dd>		04Mar97 /Don McGregor    	/changes for javadoc
 *<dd>		16Apr97 /Don McGregor    	/PrintStream passed to printValues<BR>
 *<dd>		8Dec97	/Ronan Fauglas		/changes for documentation templates + complements in documentation
 *<dd>		11Dec97	/Ronan Fauglas		/changes access methods names from "variable()" to "getVariable()"
 *<dd>		04Jan98	/Ronan Fauglas		/suppressed redundant comments for clone, printValues, etc...
 *<dd>		12Jan98	/Ronan Fauglas		/changed changeIndicator to parameterChangeIndicator.
 *<dd>		12Jan98	/Ronan Fauglas		/changed partID to articulationAttachmentID.
 *
 *<dt><b>References:</b>
 *<dd>		DIS Data Dictionary: <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/6d.htm">
 *		Articulation Parameter Record</a>
 *<dd>		DIS specification : IEEE 1278.1, Section 5.3.5, Annex A
 *
 *@see PduElement 
 *@see DetonationPdu
 *@see SerializationInterface
 *@see EntityPosition
 *@see EntityStatePdu
 */
public class ArticulationParameter extends PduElement
{
    /**
     *The identification of wether the Parameter Type Record is for an articulated 
     *  or attached part shall be designated by this field.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>Enumeration, see references below for values.
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary :<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/73.htm">
     *		Parameter Type Designator Field</a>
     *</dl>
     */
    protected   UnsignedByte    parameterTypeDesignator;        // Type of parameter (enumeration)

    /**
     *The change of any parameter for any articulated part shall be indicated by a change indicator field.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>This field shall be set to o for each exercise and sequencially incremented by one
     *  for each change in articulation parameter. In the case all possible values are exhausted, the numbers 
     *  shall be reused beginning at zero.
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary :<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/74.htm">
     *		Parameter Change Indicator Field</a>
     *<dd>	See section 4 in EBV-DOC
     *</dl>
     */
    protected   UnsignedByte    changeIndicator;                // change indicator    
    
    /**
     *The identification of the articulated part to which this articulation parameter is attached.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>This field shall contain the value 0 if the articulated part is attached directly to the entity.
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary :<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/75.htm">
     *		Articulation Attachment ID Field</a>
     *</dl>
     */
    protected   UnsignedShort   articulationAttachmentID;                         // What part we're attached to 

    /**
     *The type of parameter represented shall be specified by an enumeration.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>Enumeration, see references below for values.
     *<dt><b>Note:</b>
     *<dd>The Dis Data Dictionnary is bugged, please refer to the standard.
     *
     *<dt><b>References:</b>
     *<dd>      DIS Data Dictionary :<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/72.htm">
     *          Parameter Type Varient</a> (bugged...)
     *<dd>      DIS specification : IEEE 1278.1, Annex A
     *<dd>      See Section 4 in EBV-DOC
     *</dl>
     */
    protected   UnsignedInt     parameterType;                  // parameter type

   /**
     *This field contains the value of the Articulation parameter. 
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>The definition of the 64 bits shall be determined based on the type of parameter specified
     * in the ParameterType field. See References below for further information.
     *
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary :<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/76.htm">
     *		Articulation Parameter Record</a> (bugged...)
     *<dd>	DIS specification : IEEE 1278.1, 5.3.5.4
     *<dd>	See Section 4 in EBV-DOC
     *</dl>
     */
    protected   long            parameterValue;                 // 64 bits of parameter data 

    /**
     *Constant value--size of an Articulation Parameter WHEN WRITTEN TO THE WIRE; here :<code>sizeOf = 16 bytes</code>.
     */
    public static final int     sizeOf = 16;                    // size, in bytes, of a serialized articulation parameter */


/**
 *Default constructor--fills with zeros for all values.
 */
public ArticulationParameter()                                  // default constructor
 {
    parameterTypeDesignator = new UnsignedByte(0);
    changeIndicator         = new UnsignedByte(0);
    articulationAttachmentID                  = new UnsignedShort(0);
    parameterType           = new UnsignedInt(0);
    parameterValue          = 0;

    return;
 }

public Object clone()
{

 ArticulationParameter  newArticulationParam = new ArticulationParameter();

 newArticulationParam.setParameterTypeDesignator(this.getParameterTypeDesignator());
 newArticulationParam.setChangeIndicator(this.getChangeIndicator());
 newArticulationParam.setArticulationAttachmentID(this.getArticulationAttachmentID());
 newArticulationParam.setParameterType(this.getParameterType());
 newArticulationParam.setParameterValue(this.getParameterValue());

 return newArticulationParam;
}

/* 
 *@exception RuntimeException when IO Error occurs.
 */
public void serialize(DataOutputStream outputStream)
{
    //super.serialize(outputStream);        // No need to call super, since no ivars are in our direct superclass
    
    try
     {
        parameterTypeDesignator.serialize(outputStream);
        changeIndicator.serialize(outputStream);
        articulationAttachmentID.serialize(outputStream);
        parameterType.serialize(outputStream);
        outputStream.writeLong(parameterValue);
     }
    catch(IOException ioException)      // catch-all for all IO exceptions in the above code
     {
        throw new 
            RuntimeException("Exception in ArticulationParameter. Error serializing unit.");
     }
    
    return;
}   

/* 
 *@exception RuntimeException when IO Error occurs.
 */
public void deSerialize(DataInputStream inputStream)
{
  // no need to call super, since there are no ivars in the superclass.

  try
    {
        parameterTypeDesignator.deSerialize(inputStream);
        changeIndicator.deSerialize(inputStream);
        articulationAttachmentID.deSerialize(inputStream);
        parameterType.deSerialize(inputStream);
        parameterValue = inputStream.readLong();
     }
    catch(IOException ioException)      // catch-all for all IO exceptions in the above code
     {
        throw new 
            RuntimeException("Exception in ArticulationParameter. Error deSerializing unit.");
     }
    
    return;
}   

public int length()
{
    return sizeOf;          // articulation parameters are sizeOf bytes long, defined above
}

public void printValues(int indentLevel, PrintStream printStream)
{

    char[]  indent = new char[indentLevel];
    int     idx = 0;
    String  spacing;

    for(idx = 0; idx < indentLevel; idx++)
        indent[idx] = ' ';

    spacing = new String(indent);

    printStream.println(indent + "parameterTypeDesignator: " + parameterTypeDesignator.intValue());
    printStream.println(indent + "changeIndicator: " + changeIndicator.intValue());
    printStream.println(indent + "articulationAttachmentID: " + articulationAttachmentID.intValue());
    printStream.println(indent + "parameterType: " + parameterType.longValue());
    printStream.println(indent + "parameterValue: " + parameterValue);

    return;
}



  // accessor methods. The get() methods return a copy of the instance 
  // variable, rather than the instance variable itself.

  public UnsignedByte getParameterTypeDesignator()
  {   return (UnsignedByte)parameterTypeDesignator.clone();}

  public void setParameterTypeDesignator(UnsignedByte pParameterTypeDesignator)
  { parameterTypeDesignator = pParameterTypeDesignator; }

  public void setParameterTypeDesignator(int pParameterTypeDesignator)
  { parameterTypeDesignator = new UnsignedByte(pParameterTypeDesignator);}

  public UnsignedByte getChangeIndicator()
  { return (UnsignedByte)changeIndicator.clone();}

  public void setChangeIndicator(UnsignedByte pChangeIndicator)
  { changeIndicator = pChangeIndicator;}

  public void setChangeIndicator(int pChangeIndicator)
  { changeIndicator = new UnsignedByte(pChangeIndicator); }

  public UnsignedShort getArticulationAttachmentID()
  { return (UnsignedShort)articulationAttachmentID.clone(); }

  public void 
  setArticulationAttachmentID(UnsignedShort pArticulationAttachmentID)
  { articulationAttachmentID = pArticulationAttachmentID; }

  public void setArticulationAttachmentID(int pArticulationAttachmentID)
  { articulationAttachmentID = new UnsignedShort(pArticulationAttachmentID); }

  public UnsignedInt getParameterType()
  { return (UnsignedInt)parameterType.clone();}

  public void setParameterType(UnsignedInt pParameterType)
  { parameterType = pParameterType;}

  public void setParameterType(int pParameterType)
  { parameterType = new UnsignedInt(pParameterType);}

  public long getParameterValue()
  { return parameterValue;                              
    // Don't have to copy, since this is a basic type, not an object
  }

  public void setParameterValue(long pParameterValue)
  { parameterValue = pParameterValue;}

  /** 
  */
  public void write( org.w3c.dom.Document outDoc )
  { insertToDocument(outDoc,outDoc); }

  /** 
  */
  public void insert( org.w3c.dom.Node outNode )
  { 
    org.w3c.dom.Document outDoc = outNode.getOwnerDocument();
    insertToDocument(outDoc,outNode); 
  }

  /** 
  */
  public void 
  insertToDocument( org.w3c.dom.Document outDoc, org.w3c.dom.Node outNode )
  {

    Element ele = outDoc.createElement(iwt.DIS_TAGS.ArticulationParameter);
    ele.setAttribute(DIS_TAGS.ParameterTypeDesignator,
                              parameterTypeDesignator.toString());
    ele.setAttribute(DIS_TAGS.ChangeIndicator, changeIndicator.toString());
    ele.setAttribute(DIS_TAGS.ArticulationAttachmentID, 
                              articulationAttachmentID.toString());
    ele.setAttribute(DIS_TAGS.ParameterType, parameterType.toString());
    ele.setAttribute(DIS_TAGS.ParameterValue, Long.toString(parameterValue));

    outNode.appendChild(ele);

  } // end of insertToDocument()  

  /* 
  *@exception RuntimeException when IO Error occurs.
  */
  public void extract( org.w3c.dom.Node rt ) throws InvalidDocumentType
  {

    // no need to call super, since there are no ivars in the superclass.

    /*
    parameterTypeDesignator = new UnsignedByte(0);
    changeIndicator         = new UnsignedByte(0);
    articulationAttachmentID                  = new UnsignedShort(0);
    parameterType           = new UnsignedInt(0);
    parameterValue          = 0;
    */

    //System.out.println(" NodeName: " +rt.getNodeName());

    if( rt.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE )
    {
       //System.out.println(" tagName: " +((Element)rt).getTagName());
    if( (((Element)rt).getTagName()).equals(DIS_TAGS.ArticulationParameter) )
    {
       org.w3c.dom.NamedNodeMap attribs = rt.getAttributes();

       org.w3c.dom.Attribute a1 = (org.w3c.dom.Attribute) attribs.getNamedItem(DIS_TAGS.ParameterTypeDesignator); 
       org.w3c.dom.Attribute a2 = (org.w3c.dom.Attribute) attribs.getNamedItem(DIS_TAGS.ChangeIndicator); 
       org.w3c.dom.Attribute a3 = (org.w3c.dom.Attribute) attribs.getNamedItem(DIS_TAGS.ArticulationAttachmentID); 
       org.w3c.dom.Attribute a4 = (org.w3c.dom.Attribute) attribs.getNamedItem(DIS_TAGS.ParameterType); 
       org.w3c.dom.Attribute a5 = (org.w3c.dom.Attribute) attribs.getNamedItem(DIS_TAGS.ParameterValue); 
       String v1, v2, v3, v4, v5;

       v1 = a1.getValue();
       v2 = a2.getValue();
       v3 = a3.getValue();
       v4 = a4.getValue();
       v5 = a5.getValue();

       try
       {
         parameterTypeDesignator = new UnsignedByte(Integer.parseInt(v1)); 

       } catch(NumberFormatException nfe) {} 

       try
       {
         changeIndicator = new UnsignedByte(Integer.parseInt(v2)); 

       } catch(NumberFormatException nfe) {} 

       try
       {
         articulationAttachmentID = new UnsignedShort(Integer.parseInt(v3)); 

       } catch(NumberFormatException nfe) {} 

       try
       {
         parameterType = new UnsignedInt(Integer.parseInt(v4)); 

       } catch(NumberFormatException nfe) {} 

       try
       {
         parameterValue = Long.valueOf(v5).longValue(); 

       } catch(NumberFormatException nfe) {} 

    } else throw new InvalidDocumentType();
    } else throw new InvalidDocumentType();

  }// read   

}// end of class ArticulationParameter

