/*
 File:		CollisionPdu.java
 CVS Info:	$Id: CollisionPdu.java,v 1.2 1998/01/27 18:43:58 mcgredo Exp $
 Compiler:	jdk 1.1.6 
 */

package mil.navy.nps.dis;                        // package to which we belong

import mil.navy.nps.util.*;         // General-purpose utilities
import mil.navy.nps.disEnumerations.*;         // Enumerations for DIS

import java.lang.*;
import java.util.*;                              // utility stuff we need
import java.io.*;                                // input/output for serialization

import iwt.*;
import org.w3c.dom.*;

/**
 *@version 1.0
 *@author <a href="mailto:mcgredo@stl.nps.navy.mil"> Don McGregor</a> (<a href="http://www.stl.nps.navy.mil/~mcgredo">http://www.stl.nps.navy.mil/~mcgredo</a>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/CollisionPdu.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/CollisionPdu.java</a>
 *
 *<dt><b>Hierarchy Diagram:</b>
 *<dd><a href="images/PduClassHierarchy.jpg"><IMG SRC="images/PduClassHierarchyButton.jpg" ALIGN=ABSCENTER></a>
 *
 *<dt><b>Summary:</b>
 *<dd>Collisions between entities shall be communicated by issuing a Collision PDU.
 *
 *<dt><b>Explanation:</b>
 *<dd>   The collision pdu denotes the collision between two entities. It inherits 
 *  the header information from ProtocolDataUnit, an abstract class that contains
 *  assorted protocol information. It implements the IDs of the issuing entity, 
 *  the colliding entity, the event, and where the impact occurred with respect 
 *  to the issuing entity.<P>
 *  As with other PDUs, it knows how to serialize and deserialize itself
 *  from the wire. It also knows how to clone itself, and knows how to
 *  calculate its size when sent to the wire.<P>
 *
 *<dt><b>Note:</b>
 *<dd>   Here we don't have any padding field, because it isn't used in DIS. However, it is serialized and deSerialized
 * when written to the wire.
 * Also note that we have not created accessor methods for the nested fields.
 *
 *<dt><b>History:</b>
 *<dd>		10Sep97	/JPL			/New
 *<dd>		8Dec97	/Ronan Fauglas		/changes for documentation templates + complements in documentation
 *<dd>		11Dec97	/Ronan Fauglas		/changed access methods: thisVariable() --> getThisVariable()
 *
 *<dt><b>References:</b>
 *<dd>		DIS-Java-VRML Working Group: <a href="http://www.stl.nps.navy.mil/dis-java-vrml/"> 
 *		http://www.stl.nps.navy.mil/dis-java-vrml/</a>
 *<dd>		DIS Data Dictionary :<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/77.htm">
 *		Collision PDU</a>	
 *<dd>		DIS specification : IEEE 1278.1, Section 4.4.2.2, 5.4.3.2
 *
 *@see ProtocolDataUnit
 *@see PduElement 
 *@see SerializationInterface
 *@see EntityStatePdu
 */
public class CollisionPdu extends ProtocolDataUnit 
{

    /**
     *Issuing Entity ID - This field shall identify the entity that is issueing the Pdu
     */
    protected EntityID        issuingEntityID;   // ID of entity that noticed collision

    /**
     *Colliding Entity ID -  This field shall identify the entity that has collided with the issuing entity.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd> If the entity ID is unknown or the collision is with a terrain object, this field shall contain 
     *ENTITY_ID_UNKNOWN.
     *</dl>
     */
    protected EntityID        collidingEntityID; // ID of entity that was hit

    /**
     *Event ID - this field shall contain an identification generated by the issuing simulation application
     * to associate related collision events.
     */
    protected EventID         eventID;           // ID of event

    /**
     * Collision Type - This field shall identify the type of collision.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>Enumeration, see references below for semantics.
     *<dt><b>References:</b>
     *<dd>	see Section 10 in the EBV-doc
     *</dl>
     */
    protected UnsignedByte    collisionType;     // collision modeling info

    /**
     *Velocity - This field shall contain the velocity of the issuing entity (at the time the collision is detected). 
     *<dl>
     *<dt><b>Value:</b>
     *<dd>The velocity shall be represented in world Coordinates.
     *</dl>
     */
    protected LinearVelocity        velocity;          // Velocity of issuing entity

    /**
     *Mass -this field shall contain the mass of the issuing entity.
     *<dl>
     *<dt><b>Value:</b>
     *<dd>In kilograms.
     *</dl>     
     */
    protected float           mass;              // Mass of issuing entity

    /**
     *Location - This field shall specify the location of the collision with respect to the entity 
     *with which the issuing entity collided.
     */
    protected EntityCoordinate  location;          // x,y,z location of impact, in 32 bit coords

    /**
     *Constant value--size of an collision PDU without headder.
     *<code>sizeOf = 60 bytes</code>
     */
    public static final int   sizeOf = 60;       // size of CollisionPDU (bytes), AS WRITTEN TO WIRE

/**
 *Default constructor--fills with zeros for all values.
 */
public CollisionPdu()
{
    super.setPduType(PDUTypeField.COLLISION);

    issuingEntityID    = new EntityID();
    collidingEntityID  = new EntityID();
    eventID            = new EventID();
    collisionType      = new UnsignedByte();
    velocity     = new LinearVelocity();
    mass               = (float)0.0;
    location           = new EntityCoordinate();
}


public Object clone()
{
    /** make a deep copy, so we don't have two objects sharing pointers to the same data.
    */
 CollisionPdu  newCollisionPdu = (CollisionPdu)super.clone();

 newCollisionPdu.setIssuingEntityID(this.getIssuingEntityID());
 newCollisionPdu.setCollidingEntityID(this.getCollidingEntityID());
 newCollisionPdu.setEventID(this.getEventID());
 newCollisionPdu.setCollisionType(this.getCollisionType());
 newCollisionPdu.setLinearVelocity(this.getLinearVelocity());
 newCollisionPdu.setMass(this.getMass());
 newCollisionPdu.setLocation(this.getLocation());

 return newCollisionPdu;
}


/**
 *@exception RuntimeException when IO error occurs.
 */
public void serialize(DataOutputStream outputStream)
{
    /** write the data to an output stream in the order specified by the DIS
        standard. */

    UnsignedByte   padding = new UnsignedByte(0);

    super.serialize(outputStream);      // write out header info

    issuingEntityID.serialize(outputStream);
    collidingEntityID.serialize(outputStream);
    eventID.serialize(outputStream);
    collisionType.serialize(outputStream);
    padding.serialize(outputStream);
    velocity.serialize(outputStream);
    try {
      outputStream.writeFloat(mass);
    }
    catch(IOException ioError) {
       throw new RuntimeException("Exception in CollisionPdu. Error writing to wire.");
    }
    location.serialize(outputStream);

    return;
}


/**
 *@exception RuntimeException when IO error occurs.
 */
public void deSerialize(DataInputStream inputStream)
{
    /**
    deserialize the data from the input stream. By calling the superclass,
    the data is read into the buffer first.  Then we deserialize each of our 
    ivars to the appropriate variable stream in the order specified by the DIS
    standard. */

    UnsignedByte   padding = new UnsignedByte(0);

    super.deSerialize(inputStream);

    issuingEntityID.deSerialize(inputStream);
    collidingEntityID.deSerialize(inputStream);
    eventID.deSerialize(inputStream);
    collisionType.deSerialize(inputStream);
    padding.deSerialize(inputStream);
    velocity.deSerialize(inputStream);
    try {
       mass = inputStream.readFloat();
    }
    catch (IOException ioError) {
       throw new RuntimeException("Exception in CollisionPdu. Error reading from wire.");
    }
    location.deSerialize(inputStream);

    return;
}


public int length()
{  return sizeOf;  }


public String PDUName()
{  return new String("Collision PDU");  }


public void printValues(int indentLevel, PrintStream printStream)
{
    /** print the values of the object out, with correct level of
        indentation on the page. */

    char[]  indent = new char[indentLevel];
    String  spacing;
    int     superclassIndent = indentLevel;

    for(int ii = 0; ii < indentLevel; ++ii)
        indent[ii] = ' ';

    spacing = new String(indent);

    printStream.println();
    printStream.println("Collision PDU-");

    if(superclassIndent > 0)
      superclassIndent -= 1;

    super.printValues(superclassIndent, printStream);

    issuingEntityID.printValues(indentLevel, printStream);
    collidingEntityID.printValues(indentLevel, printStream);
    eventID.printValues(indentLevel, printStream);
    printStream.println(indent + "Collision type: "   + collisionType.intValue());
    velocity.printValues(indentLevel, printStream);
    printStream.println(indent + "Mass: "   + mass);
    location.printValues(indentLevel, printStream);

    return;
}


//Accessor methods

public EntityID getIssuingEntityID()
{  return (EntityID)issuingEntityID.clone();  }

public void setIssuingEntityID(EntityID pIssuingEntityID)
{  issuingEntityID = pIssuingEntityID;  }


public EntityID getCollidingEntityID()
{  return (EntityID)collidingEntityID.clone();  }

public void setCollidingEntityID(EntityID pCollidingEntityID)
{  collidingEntityID = pCollidingEntityID;  }


public EventID getEventID()
{  return (EventID)eventID.clone();  }

public void setEventID(EventID pEventID)
{  eventID = pEventID;  }


public UnsignedByte getCollisionType()
{  return (UnsignedByte)collisionType.clone();  }

public void setCollisionType(UnsignedByte pCollisionType)
{  collisionType = pCollisionType;  }


public LinearVelocity getLinearVelocity()
{  return (LinearVelocity)velocity.clone();  }

public void setLinearVelocity(LinearVelocity pLinearVelocity)
{  velocity = pLinearVelocity;  }

public void setLinearVelocity(float x, float y, float z)
{  velocity = new LinearVelocity(x, y, z);  }


public float getMass()
{  return mass;  }

public void setMass(float pMass)
{  mass = pMass;  }


public EntityCoordinate getLocation()
{  return (EntityCoordinate)location.clone();  }

public void setLocation(EntityCoordinate pLocation)
{  location = pLocation;  }

public void setLocation(float x, float y, float z)
{  location = new EntityCoordinate(x, y, z);  }

  /**
  */
  public void 
  write(org.w3c.dom.Document outDoc) 
  {
  }// end of write

  /**
  */
  public void
  insert(org.w3c.dom.Node outNode) 
  {
  }// end of insert

  /**
  */
  public void
  read(org.w3c.dom.Document inDoc) throws InvalidDocumentType 
  {
  }// end of insert

  /**
  */
  public void
  extract(org.w3c.dom.Node inNode)  throws InvalidDocumentType
  {
  }// end of insert

} // end of class CollisionPdu

