/*
 File:		LinearAcceleration.java
 CVS Info:	$Id: LinearAcceleration.java,v 1.2 1998/01/27 18:44:17 mcgredo Exp $
 Compiler:	jdk1.1.6
 */

package mil.navy.nps.dis;

import mil.navy.nps.util.*;         // General-purpose utilities
import mil.navy.nps.disEnumerations.*;         // Enumerations for DIS

import java.lang.*;
import java.util.*;
import java.io.*;

import iwt.*;
import org.w3c.dom.*;

/** 
 *@version 1.0
 *@author <a href="mailto:mcgredo@stl.nps.navy.mil"> Don McGregor</a> (<a href="http://www.stl.nps.navy.mil/~mcgredo">http://www.stl.nps.navy.mil/~mcgredo</a>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/LinearAcceleration.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/LinearAcceleration.java</a>
 *
 *<dt><b>Hierarchy Diagram:</b>
 *<dd><a href="images/PduClassHierarchy.jpg"><IMG SRC="images/PduClassHierarchyButton.jpg" ALIGN=ABSCENTER></a>
 *
 *<dt><b>Summary:</b>
 *<dd>Linerar Acceleration shall be represented as a vector with components in either World Coordinate System or Entities Coordinate System
 *  depending on the value in the Dead Recoking Algorithm Field.Each vector component shall represent 
 *  acceleration in meters per second squared.
 *
 *<dt><b>Explanation</b>
 *<dd>Acceleration Describes the Acceleration of an object, in x, y, z 32 bit
 *  floating point terms expressed  in world's coordinate system or entit.
 *  In fact DIS uses the vector record where an entity coordinate vector, 
 *  a linear acceleration or a linear velocity vector stands.
 *  We have a particular class for each of those, we provides us with stronger typing, hence improves security.
 *
 *<dt><b>History:</b>
 *<dd>		07Mar97 /Don McGregor    	/New
 *<dd>		16Apr97 /Don McGregor    	/PrintStream passed to printValues<BR>
 *<dd>		12Aug97 /Don Brutzman		/elaborated printValues
 *<dd>		8Dec97	/Ronan Fauglas		/changes for documentation templates + complements in documentation
 *<dd>		11Dec97	/Ronan Fauglas		/changes access methods names from "variable()" to "getVariable()"
 *
 *<dt><b>References:</b>
 *<dd>		DIS Data Dictionary: <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/33.htm">
 *			Linear Acceleration Vector Record</a>
 *<dd>		DIS specification : IEEE 1278.1, Section 5.3.33.2
 *
 *@see PduElement 
 *@see SerializationInterface
 *@see EntityCoordinate
 *@see LinearVelocity
 */

public class LinearAcceleration extends PduElement
{
    /**
     *First coordinate of acceleration along the X axis, in a cartesian coordinate system.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>In meters by second.
     *</dl>
     */
    protected float  x;     // x-position

    /**
     *Second coordinate of acceleration along the Y axis, in a cartesian coordinate system.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>In meters by second.
     *</dl>
     */
    protected float  y;     // y-position

    /**
     *Third coordinate of acceleration along the Z axis, in a cartesian coordinate system.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>In meters by second.
     *</dl>
     */
    protected float  z;     // z-position

   /**
     *Constant value--size of a Linear Acceleration record when written out; here :<code>sizeOf = 12 bytes</code>.
     */
    public final int sizeOf = 12;   // object is 12 bytes long 

/**
 *Constructs an new Acceleration Object, acceleration's value is 0.
 */
public LinearAcceleration()
{
    // default constructor

    x = 0; 
    y = 0;
    z = 0;

    return;
}

/**
 *Constructs a new Acceleration object whose coordinate values are passed in parameters.
 *
 *@param pX the first coordinate in the cartesian coordinate system
 *@param pY the second coordinate in the cartesian coordinate system
 *@param pZ the third coordinate in the cartesian coordinate system
 */
public LinearAcceleration(float pX, float pY, float pZ)
{
    x = pX;
    y = pY;
    z = pZ;

    return;
}

public Object clone()
{

 LinearAcceleration   newAcceleration = new LinearAcceleration(x, y, z);

 return newAcceleration;
}

/**
 *@exception RuntimeException when IO error occurs.
 */
public void serialize(DataOutputStream outputStream)
{
    // write out the data to an output stream. Order is important here, since
    // it needs to conform to the DIS standard.

    //super.serialize(outputStream);        // No need to call super, since no ivars are in our direct superclass
    
    try
     {
        outputStream.writeFloat(x);
        outputStream.writeFloat(y);
        outputStream.writeFloat(z);
     }
    catch (IOException ioError)
        {
            throw new 
                RuntimeException("Exception in Acceleration. Error writing to wire.");
        }
}

/**
 *@exception RuntimeException when IO error occurs.
 */
public void deSerialize(DataInputStream pInputStream)
{
    try
     {
        x = pInputStream.readFloat();
        y = pInputStream.readFloat();
        z = pInputStream.readFloat();
     }
     catch (IOException ioError)
        {
            throw new 
                RuntimeException("Exception in Acceleration. Error reading from wire.");
        }

}

public int length()
{
    return sizeOf;          // EntityTypes are this long, always.
}

public void printValues(int indentLevel, PrintStream printStream)
{
    // print the values of the object out, with correct level of
    // indentation on the page.

    char[]  indent = new char[indentLevel];
    int     idx = 0;
    String  spacing;

    for(idx = 0; idx < indentLevel; idx++)
        indent[idx] = ' ';

    spacing = new String(indent);

    printStream.println(indent + "acceleration x: "   + x);
    printStream.println(indent + "acceleration y: "   + y);
    printStream.println(indent + "acceleration z: "   + z);

    return;
}

// Accessor methods

public void setValues(float pX, float pY, float pZ)
{ x = pX; y = pY; z = pZ;}

public float getX() {   return x;}
public void setX(float pX) {x = pX;}

public float getY() {   return y;}
public void setY(float pY) {y = pY;}

public float getZ() {   return z;}
public void setZ(float pZ) {z = pZ;}

  /**
  */
  public void
  write(org.w3c.dom.Document outDoc) { insertToDocument(outDoc,outDoc); }

  /**
  */
  public void
  insert(org.w3c.dom.Node outNode) 
  { 
    org.w3c.dom.Document outDoc = outNode.getOwnerDocument();
    insertToDocument(outDoc,outNode); 

  }// end of insert()

  /**
  */
  public void
  insertToDocument(org.w3c.dom.Document outDoc, org.w3c.dom.Node outNode)
  {
    Element ele = outDoc.createElement(DIS_TAGS.LinearAcceleration);
    ele.setAttribute(DIS_TAGS.X,Float.toString(x));
    ele.setAttribute(DIS_TAGS.Y,Float.toString(y));
    ele.setAttribute(DIS_TAGS.Z,Float.toString(z));

    outNode.appendChild(ele);

  }// end of insertToDocument()

  /**
  */
  public void
  read(org.w3c.dom.Document inDoc) throws InvalidDocumentType
  {

    org.w3c.dom.NodeList nodes = inDoc.getElementsByTagName(DIS_TAGS.LinearAcceleration);
    if( nodes.getLength()<1 ) throw new InvalidDocumentType();

    org.w3c.dom.Node rt = nodes.item(0);

    extract(rt);

  }// end of read()

  /**
  */
  public void
  extract(org.w3c.dom.Node rt) throws InvalidDocumentType
  {

    if( rt!=null && rt.getNodeName().equals(DIS_TAGS.LinearAcceleration) && 
        rt.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE )
    {
       org.w3c.dom.NamedNodeMap attribs = rt.getAttributes();
       org.w3c.dom.Attribute a1 = (org.w3c.dom.Attribute) attribs.getNamedItem(DIS_TAGS.X);
       org.w3c.dom.Attribute a2 = (org.w3c.dom.Attribute) attribs.getNamedItem(DIS_TAGS.Y);
       org.w3c.dom.Attribute a3 = (org.w3c.dom.Attribute) attribs.getNamedItem(DIS_TAGS.Z);
       String v1, v2, v3;

       v1 = a1.getValue();
       v2 = a2.getValue();
       v3 = a3.getValue();

       try
       {
         x = Float.valueOf(v1).floatValue();

       } catch(NumberFormatException nfe) {}

       try
       {
         y = Float.valueOf(v2).floatValue();

       } catch(NumberFormatException nfe) {}

       try
       {
         z = Float.valueOf(v3).floatValue();

       } catch(NumberFormatException nfe) {}

    } else throw new InvalidDocumentType();

  }// end of extract()

} // end of class Acceleration
