/*
 File:		NetworkMonitor.java
 CVS Info:	$Id: NetworkMonitor.java,v 1.1.1.1 1998/01/22 21:35:14 mcgredo Exp $
 Compiler:	jdk 1.1.6 
 */

/**-----------------------------------------------------------------------------
    NetworkMonitor<P>

    <P>


----------------------------------------------------------------------------*/

package mil.navy.nps.dis;       // the package we belong to

import mil.navy.nps.util.*;         // General-purpose utilities
import mil.navy.nps.disEnumerations.*;         // Enumerations for DIS

import java.net.*;
import java.util.*;
import java.io.*;

import netscape.security.*;     // to enable netscape to read from random datagram sockets

/** 
 *@version 1.0
 *@author <a href="mailto:mcgredo@stl.nps.navy.mil">Don McGregor</a> (<a href="http://www.stl.nps.navy.mil/~mcgredo">http://www.stl.nps.navy.mil/~mcgredo</a>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/NetworkMonitor.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/NetworkMonitor.java</a>
 *
 *<dt><b>Hierarchy Diagram:</b>
 *<dd><a href="images/PduClassHierarchy.jpg"><IMG SRC="images/PduClassHierarchyButton.jpg" ALIGN=ABSCENTER></a>
 *
 *<dt><b>Summary:</b>
 *<dd>	The NetworkMonitor provides an interface for getting PDUs
 *	from the wire, decoupled from any blocking reads on the socket, and
 *	digested from the wire format into PDU objects.
 *
 *<dt><b>Explanation</b>
 *<dd>	    This class contains a thread object, which is configured to read from
 *	the wire as quickly as possible, so that as few packets fall onto the
 *	floor as possible. The reads are buffered up in a Vector object of datagrams. 
 *	When the DIS application comes asking for PDUs (in its own thread) the datagram
 *	buffer is swapped out and replaced by a new one. Once the DIS application
 *	thread has exclusive use of the vector buffer, it can translate the datagrams
 *	into full-fledged PDU objects, which are saved in another buffer. If duplicate
 *	ESPDUs exist, they can be culled either before translating, by directly accessing
 *	the datagram data, or culled after translating them into ESPDUs. Finally,
 *	the vector of PDUs is returned to the DIS application, to do with as it pleases.
 *	<P>
 *	Since there are typcially two threads operating on this data, we need to
 *	be careful about synchronized access here, so the threads don't corrupt
 *	the internal data structures.
 *	<P>
 *	This object is typically configured with the port it's going to be reading
 *	from. When instantiated, a socket on that port is opened, and a thread
 *	spawned to read it. It then begins to continuously read data.
 *	<P>
 *	Basically, this object encapsulates both a socket and a thread that reads
 *	from the socket. The advantage is that this makes the external interface
 *	very simple; as far as the consumer is concerned, he asks for and gets 
 *	PDUs that have arrived since the last time he asked, without having to 
 *	worry about blocking. Possible disadvantages include buffer overflow.
 *	Right now there's no guard against this, so if the consumer doesn't ask
 *	for PDUs, and the packets keep coming in, the datagram buffer will keep
 *	on storing them up. We don't worry about thread or socket cleanup, since
 *	the expectation is that the whole shebang will get flushed only when the 
 *	whole application quits.
 *	<P>
 *	The object implements the Runnable interface. To create a new instance
 *	of the network monitor, you need to do something like this:
 *	<P>
 *	<LISTING>
 *	NetworkMonitor aNetworkMonitor = new NetworkMonitor(8008);   // pass in socket no
 *	Thread         aThread = new Thread(aNetworkMonitor);
 *
 *	aNetworkMonitor.setThread(aThread);
 *	aThread.start(); 
 *	</LISTING>
 *	There's a potential problem also if the consumer of PDU objects asks for
 *	PDUs too quickly. If the consumer thread is in a tight loop, the datagramBuffer
 *	spends so much time locked that the reader thread can't add to it. Not
 *	sure about the best way to fix this....
 *	<P>
 *	"Monitor" is probably a bad term, since it overloads some other terminology
 *	in Java. "NetworkListener" would probably be better. Also, this should 
 *	probably be broken out so that an abstract class NetworkMonitor has two
 *	concrete sublcasses, NetworkMonitorUnicast and NetworkMonitorMulticast.
 *	<P>
 *
 *<dt><b>History:</b>
 *<dd>		15Nov96 /Don McGregor    	/New
 *<dd>		10Dec96	/Don McGregor		/Multicast changed over to 1.1 API spec
 *<dd>		7Jun97	/Don Brutzman		/Upgrading for JSAI use
 *<dd>		19Aug97	/Don McGregor		/Made "multicast" terminology consistent
 *<dd>		16May98	/Don Brutzman		/changes in documentation
 *
 *@see SimulationManager
 *@see java.lang.Object
 *@see java.lang.Thread
 *@see java.lang.Runnable
 *@see java.net.DatagramSocket
 *@see java.net.MulticastSocket
 */

public class NetworkMonitor extends Object implements Runnable
{
    // constants

    private static boolean  	DEBUG             = false;
    private static final int	MAX_DATAGRAM_SIZE = 1500;

    // ivars relating to the socket we're reading from

   boolean          usingMulticast;         // YES=> using the multicast socket NO=> using unicast socket

   int              datagramPort;           // what port we read from (used for both unicast & multicast)
   DatagramSocket   datagramSocket  = null; // the socket we read from (configured w/ datagram port)
   MulticastSocket  multicastSocket = null; // socket we read from, if we're using multicast (configured w/ datagram port & multicast address)

   String           multicastAddress;         // multicast address, doesn't work on all versions of java
   byte             ttl = 15;                 // multicast time-to-live; 15=site, 63=regional, 127=global

   // tightly coupled thread that reads from socket

   Thread           readerThread;           // Thread that continuously reads from wire

   // buffer we hold datagrams in until the DIS application asks for them

   Vector           datagramBuffer;         // holding pen for incoming datagrams

   PrivilegeManager privilegeManager;


/**
 *@exception RuntimeException when socket error occurs.
 */
public NetworkMonitor(int pDatagramPort)            // INPUT: port we read from
 {

 	if(DEBUG)
 	  System.out.println("*** NetworkMonitor: Creating new instance (this should only appear once..)");
 	  
    // Create the unicast datagram socket we read from

    usingMulticast = false;

    try
    {
    	if(DEBUG)
      System.out.println("NetworkMonitor:  try unicast new DatagramSocket(pDatagramPort)...");
        datagramSocket = new DatagramSocket(pDatagramPort);
    }
    catch (SocketException socketException)
    {
    	if(DEBUG)
      System.out.println("NetworkMonitor:  caught unicast socketException: " + socketException + "/" + socketException.getMessage ());
      throw new 
          RuntimeException("Exception in NetworkMonitor; unable to create socket in constructor");
    }
    
    datagramPort = pDatagramPort;           // not really needed, but useful info to have around, I suspect
    datagramBuffer = new Vector();

    // The thread is created and run in the startReading method

	if(DEBUG)
	  System.out.println("NetworkMonitor:  finished unicast NetworkMonitor constructor");
	  
    return;
 }

public NetworkMonitor(String pMulticastAddress, int pDatagramPort)            // INPUT: multicast Address, port we read from
{
	// NOTE: this should NEVER be called until browsers support java 1.1. Multicast
	// support just isn't there.

  // This security stuff had to be moved into the constructor, since doing the join() requires
  // privliges.

    privilegeManager = PrivilegeManager.getPrivilegeManager();
    privilegeManager.enablePrivilege("UniversalMulticast");
    privilegeManager.enablePrivilege("UniversalConnect");
    privilegeManager.enablePrivilege("UniversalAccept");
    privilegeManager.enablePrivilege("UniversalListen");

    if(DEBUG)
      System.out.println("NetworkMonitor: in multicast constructor, Address " + pMulticastAddress + " port " + pDatagramPort);

    usingMulticast = true;

    try
    {
    	if(DEBUG)
      System.out.println("NetworkMonitor:  try new MulticastSocket(pDatagramPort)...");
    	multicastSocket = new MulticastSocket(pDatagramPort);
    }
    catch (Exception socketException)
     {
       if(DEBUG)
       System.out.println("NetworkMonitor:  caught exception/unable to create multicastSocket "
           + socketException + socketException.getMessage ());

      throw new 
          RuntimeException("Exception in NetworkMonitor; unable to create multicastsocket in constructor");
     }
    if(DEBUG)
     System.out.println("NetworkMonitor:  created multicast socket");
     
    try
    {
        if(DEBUG)
          System.out.println("NetworkMonitor:  trying to join Address " + pMulticastAddress);

        multicastSocket.joinGroup(InetAddress.getByName(pMulticastAddress));
    }
    catch (Exception socketException)
     {
        throw new 
            RuntimeException("Exception in NetworkMonitor; unable to join address in constructor");
     }

    if(DEBUG)
      System.out.println("NetworkMonitor:  joined multicast address " + pMulticastAddress);
    
    datagramPort = pDatagramPort;           // not really needed, but useful info to have around, I suspect
    multicastAddress = pMulticastAddress;

    datagramBuffer = new Vector();
    ttl = 15;

	if(DEBUG)
	  System.out.println("NetworkMonitor:  finished multicast NetworkMonitor constructor");
	  
    return;
}


/**
 @@@THIS IS NETSCAPE COMMUNICATOR SPECIFIC.  It basically
 switches on network access for everything under the sun,
 so you can attach or be attached to random machines on
 the internet. enablePrivilege() turns off the priv when
 the method it's called from exits.

  You need the netscape.security package, which can be gotten
  from netscape at 
  <A HREF="http://developer.netscape.com/products/zigbert/index.html">
  http://developer.netscape.com/products/zigbert/index.html</A>
  
  Also, you'll need to either 1) sign the applet with a digital
  certificate, or, more likely, 2) change your preferences in
  Netscape to allow sleazy security. this is called activating
  codebase principals. See 
  http://developer.netscape.com/products/zigbert/index.html
*/

public void run()
{
    // loops endlessly. 

    DatagramPacket  datagramPacket;
    byte            dataBuffer[]; 

    if(DEBUG)
      	System.out.println("NetworkMonitor run (): start privilegeManager calls");

    privilegeManager = PrivilegeManager.getPrivilegeManager();
    privilegeManager.enablePrivilege("UniversalMulticast");
    privilegeManager.enablePrivilege("UniversalConnect");
    privilegeManager.enablePrivilege("UniversalAccept");
    privilegeManager.enablePrivilege("UniversalListen");

    if(DEBUG)
      	System.out.println("NetworkMonitor run (): completed privilegeManager calls");
   
    while(true)
     {
       // create a new datagram object, then read it from the wire.

      dataBuffer      = new byte[MAX_DATAGRAM_SIZE];
      datagramPacket  = new DatagramPacket(dataBuffer, dataBuffer.length);

      // read from the correct socket, either unicast or multicast.
            
      // UGLY NETSCAPE BUG ALERT: In netscape 4.02-4.04, at least, there is
      // an error when reading from datagram (UDP) sockets. An exception will
      // be thrown when a packet actually arrives. the workaround is to read
      // from the socket _again_, in the catch statement.

      try
       {
        if(usingMulticast)
          multicastSocket.receive(datagramPacket);
        else
          datagramSocket.receive(datagramPacket); 


       } // end of try
     catch (IOException ioError)
     {
      try
      {
        if(usingMulticast)
          multicastSocket.receive(datagramPacket);
        else
          datagramSocket.receive(datagramPacket);
      }  // end of second-level try
      catch(IOException io2error)
      {
        throw new 
            RuntimeException("Exception in NetworkMonitor. Error reading from datagram socket.");
      } // end of second-level catch
     } // end of first-level catch
  

     // we have to be careful about access to the datagram holding pen
     synchronized(datagramBuffer)
      {
       datagramBuffer.addElement(datagramPacket);
      }
    }   // end of while

}

public Vector receivedPdus()
{
    // Returns a Vector of all the PDUs received since we last asked.
    // Creates a new datagram buffer for the read thread, then takes the
    // old datagram buffer for its own nefarious purposes.
    
    Vector  localDatagramBuffer;                // just a ptr to keep the old buffer from being GC'd.
    Vector  pduBuffer = new Vector();           // vector to hold inflated PDU objects
    int     idx = 0;
    Vector  newDatagramBuffer;
    Enumeration datagramEnumeration;

    // Try to make the changeover fast.

    newDatagramBuffer = new Vector(); 

    // don't want to stomp on this while another thread is operating on it.

    if(DEBUG)
      	System.out.println("NetworkMonitor receivedPdus (): ready for synchronized ()...");

    if(DEBUG)
    {
      System.out.println("NetworkMonitor receivedPdus (): datagramBuffer size is " + datagramBuffer.size());
    }
    
    synchronized(datagramBuffer)
    {
    	if(DEBUG)
         {
          System.out.println("NetworkMonitor receivedPdus (): datagramBuffer size is " + datagramBuffer.size());
          System.out.println("NetworkMonitor receivedPdus (): starting swap");
         }
         
    	localDatagramBuffer = datagramBuffer;
        datagramBuffer = newDatagramBuffer; // a fast swap, with minimal locked time

        if(DEBUG)
          {
           System.out.println("NetworkMonitor receivedPdus (): datagramBuffer size is " + datagramBuffer.size());
          }
    }

    if(DEBUG)
      	System.out.println("NetworkMonitor receivedPdus (): synchronized () complete");

     // translate all the datagrams into PDUs.

    if(DEBUG)
    {
      System.out.println("NetworkMonitor receivedPdus (): got buffer of " + localDatagramBuffer.size() + " datagrams, translating into PDUs");
    }
    
    datagramEnumeration = localDatagramBuffer.elements();
    while(datagramEnumeration.hasMoreElements())
    {
        DatagramPacket      aDatagram = (DatagramPacket)datagramEnumeration.nextElement();
        ProtocolDataUnit    aPdu;

        aPdu = ProtocolDataUnit.datagramToPdu(aDatagram);
        if(aPdu != null)
            pduBuffer.addElement(aPdu);
    }
    if(DEBUG)
      	System.out.println("NetworkMonitor receivedPdus (): pduBuffer.size () = " + pduBuffer.size ());

    return pduBuffer;
}

/** 
Set the multicast socket time-to-live. 15=> local distribution, 63=>regional
distribution, 127=>global distribution. It seems that there is a bug under NS
4.04 (surprise, surprise) so that setTTL on the mcast socket throws an exception,
so you probably shouldn't use this yet.
*/

public void setTTL(byte pTTL)
{
  if(usingMulticast)  // only makes sense on mcast sockets
  {
    ttl = pTTL;
    try
    {
       multicastSocket.setTTL(ttl);
    }
    catch(IOException ioe)
    {
      throw new RuntimeException("cannot change TTL on multicast socket");
    } // end catch
  }   // end usingMulticast
}

public void sendPDUMcast(ProtocolDataUnit pPdu)
{
  if(usingMulticast)
  {
    this.sendPDU(pPdu, multicastAddress, datagramPort);
  }
}

public synchronized void sendPDU(ProtocolDataUnit pdu,
                                   String pDestinationHost,
                                   int    pDestinationSocket)
{
	ByteArrayOutputStream oStream = null;
	DataOutputStream      dataOutputStream = null;
	DatagramPacket        dgramPacket = null;
	InetAddress           destHostAddress = null;

    if(DEBUG)
     System.out.println("NetworkMonitor: in sendPDU with destination:" + pDestinationHost +
                        " on port " + pDestinationSocket);
     
    privilegeManager = PrivilegeManager.getPrivilegeManager();
    privilegeManager.enablePrivilege("UniversalMulticast");
    privilegeManager.enablePrivilege("UniversalConnect");
    privilegeManager.enablePrivilege("UniversalAccept");
    privilegeManager.enablePrivilege("UniversalListen");

	try
	 {
	 	destHostAddress = InetAddress.getByName(pDestinationHost);
	 }
	catch(UnknownHostException e)
	 {
	 	System.out.println("cannot resolve address for destination " + pDestinationHost + ", packet not sent");
	 }

	oStream = new ByteArrayOutputStream();
	dataOutputStream = new DataOutputStream(oStream);

	if(DEBUG)
	  System.out.println("Preparing to serialize a PDU");

	pdu.serialize(dataOutputStream);

  if(usingMulticast)
    dgramPacket = new DatagramPacket(oStream.toByteArray(),
	                                 oStream.size(),
	                                 destHostAddress,
	                                 datagramPort);
   else
	  dgramPacket = new DatagramPacket(oStream.toByteArray(),
	                                 oStream.size(),
	                                 destHostAddress,
	                                 pDestinationSocket);

	if(DEBUG)
	  System.out.println("created datagram packet for sending");

	try
	{
    if(usingMulticast)
      multicastSocket.send(dgramPacket);
     else
		  datagramSocket.send(dgramPacket);
	}
	catch(IOException e)
	{
		System.out.println("failed to send packet");
	}

  return;
}
	                                 

public void setThread (Thread pThread)   // INPUT: thread that runs us
{
  // setDaemon sets it to be a "background" thread. The application will
  // run until all non-background threads have run to completion. Setting
  // this to be a daemon thread allows the app to terminate even if this
  // thread is still running, which is fine.

  // We need a pointer to the thread around so we can do suspend(), resume(),
  // etc on the thread.
  
  if(DEBUG) System.out.println("NetworkMonitor setThread ():  start");
  
  readerThread = pThread;

//  setDaemon causes an exception when used in combination with Communicator
//    _and_ CosmoPlayer
//  readerThread.setDaemon(true);

  if(DEBUG) System.out.println("NetworkMonitor setThread ():  complete");
  
  return;
}


public void suspendReading()
{
  // Suspends the thread, which stops reading packets. To start up the thread
  // again, call resumeReading(). Note that there may be some stuff still
  // saved up in the incoming buffer, which will still be there if the thread
  // is resumed.

  try
  {
    readerThread.suspend();
  }
  catch(SecurityException secException)
  {
    throw new 
            RuntimeException("Exception in NetworkMonitor; no right to suspend thread");
  }
}

public void resumeReading()
{
  try
  {
    readerThread.resume();
  }
  catch(SecurityException secException)
  {
    throw new 
            RuntimeException("Exception in NetworkMonitor; no right to suspend thread");
  }
}

/**
  Finalize method--used to clean up any sockets that are still open
*/

protected void finalize() // throws Throwable
{
  if(datagramSocket != null)
    datagramSocket.close();

  if(multicastSocket != null)
    multicastSocket.close();

}
  

} // end of class NetworkMonitor

