/*
 File:		ProtocolDataUnit.java
 CVS Info:	$Id: ProtocolDataUnit.java,v 1.7 1998/02/03 17:29:44 mcgredo Exp $
 Compiler:	jdk 1.1.6
 */

package mil.navy.nps.dis;

import mil.navy.nps.util.*;         // General-purpose utilities
import mil.navy.nps.disEnumerations.*;         // Enumerations for DIS

import java.io.*;
import java.net.*;
import java.util.*;

import iwt.*;
import iwt.domUtil.*;
import org.w3c.dom.*;


/**
 *@version 1.0
 *@author <a href="mailto:mcgredo@stl.nps.navy.mil"> Don McGregor</a> (<a href="http://www.stl.nps.navy.mil/~mcgredo">http://www.stl.nps.navy.mil/~mcgredo</a>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/ProtocolDataUnit.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/ProtocolDataUnit.java</a>
 *
 *<dt><b>Hierarchy Diagram:</b>
 *<dd><a href="images/PduClassHierarchy.jpg"><IMG SRC="images/PduClassHierarchyButton.jpg" ALIGN=ABSCENTER></a>
 *
 *<dt><b>Summary:</b>
 *<dd>This is an abstract class that actual PDUs inherit from. The ProtocolDataUnit
 *  includes the header information that is present in every PDU. Centralizing
 *  this information in one class prevents us from having to re-declare it in
 *  every class.
 *
 *<dt><b>Explanation:</b>
 *<dd>
 *
 *  This class in turn inherits from PduElement, the abstract class that defines
 *  a high-level interface to the "things" that make up a PDU.<P>
 *
 *  I generally declare instance variables to be protected, which means that
 *  they are directly accessible from this class and all subclasses. Those
 *  who believe in full-jackboot mode will want them declared private, so 
 *  subclasses can't access them either.<P>
 *
 *  The accessor methods (getProtocolVersion et al) are declared public, so
 *  that anyone, including those outside this package, can access them. You
 *  should always go through accessor methods when setting values inside an
 *  object. Direct access from outside the class can cause mysterious errors
 *  that are very hard to track down.<P>
 *
 *  Note that, due to garbage collection, we don't have to worry about the
 *  status of orphaned instance variables. For example, this operation will
 *  cause memory leaks in C++ or Obj-C:<P>
 *<pre>
 *    protocolVersion = newProtocolVersion;
 *</pre>
 *  In C++, this operation would leave an orphaned object; the _old_ object
 *  that was held by protocolVersion would have no valid pointers to it, but
 *  would still take up memory. With GC, the old, orphaned protocolVersion's
 *  memory will be scavanged and returned to use.<P>
 *
 *  Accessor methods return a copy of the thing they're getting, not the
 *  actual thing. This prevents violations of encapsulation. If this weren't
 *  the case, we might see something like this:<P>
 *<code><pre>
 *myLength = aPdu.getLength();
 *myLength.addOne();
 *</pre></code>
 *  this would return another pointer to the same object contained inside of
 *  aPdu. addOne() would modify the value inside of aPdu, a violation of
 *  encapsulation. So instead we implement getLength() like this:<P>
 *
 *<code><pre>
 *public UnsignedByte getLength()
 *{ return (UnsignedByte)length.clone();
 *}
 *</pre></code>
 *
 *  this creates an identical copy of length and returns that. The calling
 *  method can modify that to its heart's content without affecting aPdu.
 *  For the same reason, the clone() method should make copies of all the
 *  instance variables. Otherwise, the "new" object will have pointers 
 *  shared with the old object. The clone() operation also helps prevent
 *  big dependency meshes for GC.<P>
 *
 *  Serialization should make a call to the superclass before it does its own
 *  instance variables. <P>
 *
 *  Note that we do not include the padding variables in the instance variables
 *  of objects. Padding is used only in the serialized, external representation
 *  of a PDU. The user would have no reason to do anything with padding in
 *  an object. Since it is only an artifact of serialization, references to
 *  padding are limited to the serialization/deserialization methods.<P>
 *
 *  Also, the length, a field in the header portion, is calculated on the fly
 *  rather than saved and set. The header has a fixed length, 18 bytes. To find
 *  the size of a PDU, implement the length() method in all the subclasses. This
 *  should make a call to super, do any local calculations required, then return
 *  the correct number. Your code might look like :<P>
 *
 *<code><pre>
 *    public int length() {
 *	return super.length() + 110 + aVector.size * anElement.length();
 *    }
 *</pre></code><P>
 *
 *  This adds up the length of the superclass, such as the PDU header,
 *  the basic length of the subclass (such as the entity state PDU), and
 *  adds the length of any variable number of parameters attached to the
 *  PDU.<P>
 *
 *<dt><b>History:</b>
 *<dd>		10Oct96	/Don McGregor		/New
 *<dd>		15Oct96	/Don McGregor		/setter methods with "primitive" types, minor changes as
 *                  				suggested in class.
 *<dd>		28Oct96	/Don McGregor		/changed length() implementation
 *<dd>		17Nov96	/Don McGregor		/retrofitted in EntityID object for site, application, entity fields,
 *						kept old external interface, added new interface for setting and
 *                  				getting entityID as one object
 *<dd>		18Nov96	/Don McGregor		/Moved entityID into subclasses of ProtocolDataUnit
 *<dd>		10Mar97	/Don McGregor   	/changes for javadoc
 *<dd>		16Apr97 /Don McGregor    	/PrintStream passed to printValues
 *<dd>		8Dec97  /Ronan Fauglas		/changes for documentation templates + complements in documentation<br>
 *<dd>		11Dec97	/Ronan Fauglas		/changed access methods to class variables to "getVariable"
 *
 *<dt><b>References:</b>
 *<dd>		DIS Data Dictionary :<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/7.htm">
 *		PDU Header Record</a>	
 *<dd>		DIS-Java-VRML Working Group: <a href="http://www.stl.nps.navy.mil/dis-java-vrml/"> 
 *		http://www.stl.nps.navy.mil/dis-java-vrml/</a>
 *<dd>		DIS specification : IEEE 1278.1, 4.4.1
 *
 *@see PduElement 
 *@see SerializationInterface
 */
public abstract class ProtocolDataUnit extends PduElement implements XMLazable
{
    /**
     *Debugging flag; when set true, turns on verbose diagnostic, statements that prints in the java console.
     *It is by default set to TRUE.
     */
    public static final boolean DEBUG  = false;

    private static long  simulationStartTime = 0;   // number of ms since 1970 until midnight on start of simulation

    // little known feature: static initialization block. Used to initalize 
    // simulationStartTime to the number of ms from 1970 until 12:00 midnight
    // on the day the simulation starts up. This uses older, deprecated methods
    // in Date. This is for backwards compatibility with java 1.02. Eventually,
    // the Java 1.1 Calendar object should be used instead.
 
    static
    {  
      Date  currentDate = new Date();
      Date  midnight;
      int   year, month, date;

      year  = currentDate.getYear();
      month = currentDate.getMonth();
      date  = currentDate.getDate();

      midnight = new Date(year, month, date); // initialized to midnight
      simulationStartTime = midnight.getTime();
    }

    // PDU header information

    /**
     *Current version of the protocol being used.   
     *It is now 2.0 which should be mapped to <code> protocolVersion=5 </code>
     *No default value is set.
     *
     *<dl>
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary:<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/5.htm"> Protocol Version Field</a>
     *<dd>	DIS specification : IEEE 1278.1, 4.4.1.1
     *</dl>
     */
    protected UnsignedByte      protocolVersion;    // What version of protocol being used

    /**
     *Exercise Identifier Field.
     *Exercise Identification shall be unique to each exercise being conducted simultaneously on the same
     *communications medium. No default value.
     *   
     *<dl>
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary :<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/6.htm">
     *		Exercise Identifier Field</a>
     *<dd>	DIS specification : IEEE 1278.1, 4.4.1.2
     *</dl>
     */
    protected UnsignedByte      exerciseID;         // ID of exercise we're participating in

    /**
     *Type of PDU (entity state, message, etc).
     *   
     *<dl>
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary: <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/8.htm">
     *		PDU Type Field</a>
     *<dd>	DIS specification : IEEE 1278.1, 4.4.1.3
     *</dl>
     *
     *@see mil.navy.nps.disEnumerations.PDUTypeField
     */
    protected UnsignedByte      pduType; 

    /**
     *Indicates the family of protocols which the PDU belongs to. 
     *
     *<dl>
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary: <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/9.htm">
     *		Protocol Family Field</a>
     *<dd>	DIS specification : IEEE 1278.1, 4.4.1.3
     *</dl>
     */
    protected UnsignedByte      protocolFamily;     // protocol family

    /**
     *Time Stamp, either relative or UTC
     *This field shall specify the time which the data in the PDU is valid.
     *
     *<dl>
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary: <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/a.htm">
     *		Time Stamp Field</a>
     *<dd>	DIS specification : IEEE 1278.1, 4.4.1
     *
     *<dt><b>Note:</b>
     *<dd>TimeStamp is initialized on creation of the Pdus.
     *</dl>
     */
    protected UnsignedInt       timestamp;          // timestamp, either relative or UTC
                                                    // length of entire PDU (not an ivar, calculated on the fly)
                                                    // padding to bring up to 32-bit word boundary (not an ivar)

    /**
     *Constant value--size of a PDU header when written out; 
     *<code>sizeOf = 12 bytes</code>.
     */
    public static final int     sizeOf = 12;        // size of header when written out

/**
 *Constructor for a default protocol data unit. No field is initialized.
 */
public ProtocolDataUnit()
{

    protocolVersion = new UnsignedByte();

    exerciseID      = new UnsignedByte();

    pduType         = new UnsignedByte();

    protocolFamily  = new UnsignedByte();

//    timestamp       = new UnsignedInt();

    // Timestamp, set to VRML DEFINITION OF TIME

    this.makeTimestampCurrent();

    return;
}

public Object clone()
{
 ProtocolDataUnit  newPdu = (ProtocolDataUnit)super.clone();
 
 newPdu.setProtocolVersion(this.getProtocolVersion());
 newPdu.setExerciseID(this.getExerciseID());
 newPdu.setPduType(this.getPduType());
 newPdu.setProtocolFamily(this.getProtocolFamily());
 newPdu.makeTimestampCurrent();

 return newPdu;
}

public void printValues(int indentLevel, PrintStream printStream)
{
    // print the values of the object out, with correct level of
    // indentation on the page.

    char[]  indent = new char[indentLevel];
    int     idx = 0;
    String  spacing;

    for(idx = 0; idx < indentLevel; idx++)
        indent[idx] = ' ';

    spacing = new String(indent);

    printStream.println(indent + "protocolVersion: " + protocolVersion.intValue());
    printStream.println(indent + "exerciseID: " + exerciseID.intValue());
    printStream.println(indent + "pduType: " + pduType.intValue());
    printStream.println(indent + "protocolFamily: " + protocolFamily.intValue());
    printStream.println(indent + "timestamp: " + timestamp.longValue());
    printStream.println(indent + "length: " + this.length());

    return;
}

/**
 *This methods sets the  Timestamp to the current Time.
 *<dl>
 *<dt><b>Explanation:</b>
 *<dd>Sets the timestamp field to a current value. <b>THIS USES THE VRML DEFINITION
 *OF TIME RATHER THAN THE DIS DEFINITION OF TIME</b>. VRML uses a double to hold
 *the number of seconds since January 1, 1970. That takes 64 bits, though, and
 *we have only 32 in the PDU. We get around this, sort of, by <b>adopting the
 *convention of having the start time equal to midnight on the day the simulation
 *started up</b>. With 32 bits, we can run for about 50 days before the number of
 *milliseconds in the field rolls over. When we return the VRML timestamp,
 *we can just add the start time (saved in a class variable) to the timestamp
 *and cast it to a double.
 *</dl>
 */
public void makeTimestampCurrent()

{

  long    currentTime = System.currentTimeMillis();    // ms since 1970



  currentTime = currentTime - simulationStartTime;     // ms since midnight on day sim started

  timestamp           = new UnsignedInt(currentTime); // instantiate new timestamp. 

                                                        

  return;

}

/**

Returns the timestamp, INTERPRETED IN THE VRML DEFINITION OF TIME. VRML uses the

number of seconds since January 1, 1970, in a double, as the definition of absolute

time. DIS uses an unsigned int with some absurdly irrelevant definition of time.

*/

public double getVRMLTimestamp()

{

  long  pduTimestamp;



  pduTimestamp = timestamp.longValue();       // retrieve value as a long

  pduTimestamp = simulationStartTime + pduTimestamp;

  return (double)pduTimestamp;

}

/**

  sets the simulation start time

*/



public static void setSimulationStartTime(long pStartTime)

{

  simulationStartTime = pStartTime;

}

/**
 *Returns the name of the PDU--eg, Entity State, Fire, Comment, etc-- as a String.
 *
 *<dl>
 *<dt><b>References:</b>
 *<dd>		See <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/index.htm">
 *		PDU Name</a>
 *</dl>
 *
 *@return the name of the PDU as a <code>String</code>
 */
public abstract String PDUName();

// Accessor methods

public UnsignedByte getProtocolVersion()
{ return (UnsignedByte)protocolVersion.clone();
}

public void setProtocolVersion(UnsignedByte pProtocolVersion)
{ protocolVersion = pProtocolVersion;
}

public void setProtocolVersion(int pProtocolVersion)
{ protocolVersion = new UnsignedByte(pProtocolVersion);
}

public UnsignedByte getExerciseID()
{ return (UnsignedByte)exerciseID.clone();
}

public void setExerciseID(UnsignedByte pExerciseID)
{ exerciseID = pExerciseID;
}

public void setExerciseID(int pExerciseID)
{ exerciseID = new UnsignedByte(pExerciseID);
}

public UnsignedByte getPduType()
{ return (UnsignedByte)pduType.clone();
}

public void setPduType(UnsignedByte pPduType)
{ pduType = pPduType;
}

public void setPduType(short pPduType)
{ pduType = new UnsignedByte(pPduType);
}

public UnsignedByte getProtocolFamily()
{ return (UnsignedByte)protocolFamily.clone();
}

public void setProtocolFamily(UnsignedByte pProtocolFamily)
{ protocolFamily = pProtocolFamily;
}

public void setProtocolFamily(int pProtocolFamily)
{ protocolFamily = new UnsignedByte(pProtocolFamily);
}

public UnsignedInt getTimestamp()
{ return (UnsignedInt)timestamp.clone();
}

public void setTimestamp(UnsignedInt pTimestamp)
{ timestamp = pTimestamp;
}

public void setTimestamp(long pTimestamp)
{ timestamp = new UnsignedInt(pTimestamp);
}

/**
 *Returns the length of the PDU header.
 *
 *@return the length of the PDU header (currently 12 bytes).
 */
public int length()
{ return ProtocolDataUnit.sizeOf;           // we're always this long
}

public void  serialize(DataOutputStream outputStream)
{
    UnsignedShort   padding = new UnsignedShort(0);             // padding to bring pdu up to word boundary
    UnsignedShort   length  = new UnsignedShort(this.length()); // how long this PDU is, calculated on the fly

    if(DEBUG)
      System.out.println("serializing in PDU");

    protocolVersion.serialize(outputStream);
    exerciseID.serialize(outputStream);
    pduType.serialize(outputStream);
    protocolFamily.serialize(outputStream);
    timestamp.serialize(outputStream);
    length.serialize(outputStream);
    padding.serialize(outputStream);

    if(DEBUG)
     System.out.println("exiting serialize in ProtocolDataUnit");

    return;
}

public void  deSerialize(DataInputStream inputStream)
{
    // Read from a stream. The order in which variables
    // are read is significant.

    UnsignedShort   padding = new UnsignedShort(0);     // padding--read just to be thrown away
    UnsignedShort   length  = new UnsignedShort(0);     // length--read just to be thrown away

    protocolVersion.deSerialize(inputStream);
    exerciseID.deSerialize(inputStream);
    pduType.deSerialize(inputStream);
    protocolFamily.deSerialize(inputStream);
    timestamp.deSerialize(inputStream);
    length.deSerialize(inputStream);			// Nothing is done with this value
    padding.deSerialize(inputStream);                   // Nothing is done with this value

    /*
    System.out.println(" protocolVersion .: "+protocolVersion.intValue());
    System.out.println(" exerciseID ......: "+exerciseID.intValue());
    System.out.println(" pduType .........: "+pduType.intValue());
    System.out.println(" protocolFamily ..: "+protocolFamily.intValue());
    System.out.println(" timestamp .......: "+timestamp);
    System.out.println(" length ..........: "+length.intValue());
    System.out.println(" padding .........: "+padding.intValue());
    System.out.println(" sizeOf:"+sizeOf);
    */
 
    return;
}

/**
 *Returns a PDU completely read in from the Datagram in parameter.
 *
 *<dl>
 *<dt><b>Explanation:</b>
 *<dd>
 *  This can't be a constructor, since we don't know what type of object
 *  we're going to instantiate before it fires. This is a static (class)
 *  method.
 *
 *  This gets a bit convoluted, but it makes the external interface nice.
 *  We have to decide what type of PDU to instantiate, but we don't know
 *  the PDU type until we've read some of it. So we read the first few
 *  bytes of the byte array in the datagram packet, and find the PDU type.
 *  Then we reset the stream, and tell the new object to deserialize itself.
 *  It does this by making a call to super.deserialize as the first executable
 *  statement of the deserialize method. This causes the data to be read
 *  in the correct order--the data for the PDU header is at the front of the
 *  datagram, and that's the first data that's deserialized. Then the subclass
 *  deserialize method fires, and that deserializes the next data in the 
 *  stream, corresponding to that subclass.
 *  
 *  The process looks something like this:
 *  <code><pre>
 *          ProtocolDataUnit
 *                 |
 *                 ^
 *           EntityStatePDU
 *  </pre></code>
 *  The deserialize message goes first to the EntityStatePdu. The ESPDU
 *  immediately calls the deserialize method in its superclass, ProtocolDataUnit.
 *  PDU deserializes the first few instance variables, then returns. The ESPDU
 *  then deserializes its ivars. 
 *  Prints out some information during exectuion if debugging flag is set.
 *</dl>
 *
 *@param pDatagramPacket The datagram packet read from the wire
 *@return the PDU as identified in the datagam (with default initialization)
 *@throws RuntimeException if an IO error occurs.
 */
public static ProtocolDataUnit datagramToPdu(DatagramPacket pDatagramPacket)
{
    ProtocolDataUnit        newPdu;                         // The pdu we return from the datagram packet
    byte                    initialBuf[];                   // data from datagram packet
    ByteArrayInputStream    initialByteStream;              // byte stream
    DataInputStream         initialDataStream;              // data stream
    UnsignedByte            pduVersion = new UnsignedByte();// pdu version

  if(DEBUG)
      System.out.println("ProtocolDataUnit: translating a datagram into a PDU");
      
    // Turn the byte array into an input data stream, so we can use nice, high-level
    // functions to read what's in it.

    initialBuf = pDatagramPacket.getData();                         // get data portion of Datagram
    initialByteStream = new ByteArrayInputStream(initialBuf);       // wrap a byte stream around it
    initialDataStream = new DataInputStream(initialByteStream);     // wrap a data stream around _that_

    // The pdu version is the third byte in the stream. Read the first two just
    // to throw them away.

    try
     {
        initialDataStream.skipBytes(2);                 // Skip first two bytes to get to pdu type
        pduVersion.deSerialize(initialDataStream);      // read the PDU type, keep the results
     }
    catch (IOException ioError)
     {
        throw new 
            RuntimeException("Exception in ProtocolDataUnit. Error instantiating from datagram.");
     }
    
    try
     {
        initialDataStream.reset();          // reset the stream to the start
     }
    catch(IOException ioException)
     {
        throw new 
            RuntimeException("Exception in ProtocolDataUnit. Error resetting stream.");
     }

    // We need to degrade gracefully if this is an unrecognized PDU...
    // we'll probably see a lot of pdus that we haven't implemented yet.

    if(DEBUG)
      System.out.println("Got PDU of type " + pduVersion);

    switch(pduVersion.shortValue())
     {
        // entity state pdu
        case PDUTypeField.ENTITYSTATE:      // do EntityStatePdu 

                //printStream.println("got a packet, instantiating an ESPDU");
                newPdu = new EntityStatePdu();
                break;


        // Fire pdu
        case PDUTypeField.FIRE:              
                if(DEBUG)
                  System.out.println("got a packet, instantiating a FIRE PDU");
                newPdu = new FirePdu();
                break;
        
        // detonation PDU, the fire PDU's frequent companion.
        case PDUTypeField.DETONATION:
            //printStream.println("got a packet, instantiating a DETONATATION PDU");
            newPdu = new DetonationPdu();
            break;

        // Message PDU, same as CommentPDU, I THINK.
        case PDUTypeField.MESSAGE:   
           // printStream.println("got a message PDU");
            newPdu = new CommentPdu();
            break;

        //case PDUTypeField.COLLISION:           return;         // do collision pdu here
    

        default:
            {
                // unrecognized PDU type. 
              System.out.println("Unrecognized PDU type received in packet:"  + pduVersion.shortValue());
                return null;
            }
    }

    newPdu.deSerialize(initialDataStream);
    //newPdu.printValues(7);

  if(DEBUG)
      System.out.println("ProtocolDataUnit: translated a datagram into a PDU");
      
    return newPdu;      // may be null if PDU type not recognized
}

  public void  writeHeader(org.w3c.dom.Document outDoc)
  {

    if(DEBUG) System.out.println("serializing in PDU");

    Element ele = outDoc.createElement(iwt.DIS_TAGS.DIS_PDU_Header);
    ele.setAttribute(iwt.DIS_TAGS.ProtocolVersion,protocolVersion.toString());
    ele.setAttribute(iwt.DIS_TAGS.ExerciseID,exerciseID.toString());
    ele.setAttribute(iwt.DIS_TAGS.PduType,pduType.toString());
    ele.setAttribute(iwt.DIS_TAGS.ProtocolFamily,protocolFamily.toString());
    ele.setAttribute(iwt.DIS_TAGS.TimeStamp,timestamp.toString());

    outDoc.appendChild(ele);
    printDOM.printElement(ele,System.out);

  }

  public void  insertHeader(org.w3c.dom.Node outNode)
  {

    if(DEBUG) System.out.println("serializing in PDU");
    Document outDoc = outNode.getOwnerDocument();

    Element ele = outDoc.createElement(iwt.DIS_TAGS.DIS_PDU_Header);
    ele.setAttribute(iwt.DIS_TAGS.ProtocolVersion,protocolVersion.toString());
    ele.setAttribute(iwt.DIS_TAGS.ExerciseID,exerciseID.toString());
    ele.setAttribute(iwt.DIS_TAGS.PduType,pduType.toString());
    ele.setAttribute(iwt.DIS_TAGS.ProtocolFamily,protocolFamily.toString());
    ele.setAttribute(iwt.DIS_TAGS.TimeStamp,timestamp.toString());

    outNode.appendChild(ele);
    printDOM.printElement(ele,System.out);

  }

  /**
  */
  public void  readHeader(org.w3c.dom.Document inDoc)
  {

    org.w3c.dom.NodeList nodes = inDoc.getElementsByTagName(iwt.DIS_TAGS.DIS_PDU_Header);
    if( nodes.getLength()<1 ) throw new iwt.InvalidDocumentType();

    org.w3c.dom.Node rt = nodes.item(0);

    extractHeader(rt);

  }// end of readHeader()

  /**
  */
  public void  extractHeader(org.w3c.dom.Node rt)
  {
    if( rt.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE &&
        rt.getNodeName().equals(iwt.DIS_TAGS.DIS_PDU_Header) )
    {
       org.w3c.dom.NamedNodeMap attribs = rt.getAttributes();
       org.w3c.dom.Attribute a1 = (org.w3c.dom.Attribute) attribs.getNamedItem(iwt.DIS_TAGS.ProtocolVersion);
       org.w3c.dom.Attribute a2 = (org.w3c.dom.Attribute) attribs.getNamedItem(iwt.DIS_TAGS.ExerciseID);
       org.w3c.dom.Attribute a3 = (org.w3c.dom.Attribute) attribs.getNamedItem(iwt.DIS_TAGS.PduType);
       org.w3c.dom.Attribute a4 = (org.w3c.dom.Attribute) attribs.getNamedItem(iwt.DIS_TAGS.ProtocolFamily);
       org.w3c.dom.Attribute a5 = (org.w3c.dom.Attribute) attribs.getNamedItem(iwt.DIS_TAGS.TimeStamp);
       String v1, v2, v3, v4, v5;

       v1 = a1.getValue();
       v2 = a2.getValue();
       v3 = a3.getValue();
       v4 = a4.getValue();
       v5 = a5.getValue();

       try
       {
         protocolVersion = new UnsignedByte(Integer.parseInt(v1));

       } catch(NumberFormatException nfe) {}

       try
       {
         exerciseID = new UnsignedByte(Integer.parseInt(v2));

       } catch(NumberFormatException nfe) {}

       try
       {
         pduType = new UnsignedByte(Integer.parseInt(v3));

       } catch(NumberFormatException nfe) {}

       try
       {
         protocolFamily = new UnsignedByte(Integer.parseInt(v4));

       } catch(NumberFormatException nfe) {}

       try
       {
         timestamp = new UnsignedInt(Integer.parseInt(v5));

       } catch(NumberFormatException nfe) {}

    } else throw new iwt.InvalidDocumentType();

  }// end of extract()

}   // end of class ProtocolDataUnit


