/*
 File:		SimulationManagement.java
 CVS Info:	$Id: SimulationManagementPdu.java,v 1.2 1998/01/27 18:44:27 mcgredo Exp $
 Compiler:	jdk 1.1.6 
 */

package mil.navy.nps.dis;                                       // package to which we belong

import mil.navy.nps.util.*;         // General-purpose utilities
import mil.navy.nps.disEnumerations.*;         // Enumerations for DIS

import java.util.*;                                             // utility stuff we need
import java.io.*;                                               // input/output for serialization

/**
 *@version 1.0
 *@author Antonio Alexandre Rua (<A HREF="http://www.garfield.fe.up.pt/~alexrua">http://www.garfield.fe.up.pt/~alexrua</A>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/CommentPdu.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/CommentPdu.java</a>
 *
 *<dt><b>Hierarchy Diagram:</b>
 *<dd><a href="images/PduClassHierarchy.jpg"><IMG SRC="images/PduClassHierarchyButton.jpg" ALIGN=ABSCENTER></a>
 *
 *<dt><b>Summary:</b>
 *<dd>The Simulation Management PDU is an abstract class, for all simulation
 *  management family PDUs. 
 *
 *<dt><b>Explanation:</b>
 *<dd>It encapsulates the specific header for the simulation Management PDUs 
 *  which includes the PDU header, the originating entity identity, and
 *  The advantage is that, we define methods to deal with this header, 
 *  and all the classes inherit from it, avoiding
 *  rewriting them a lot of times.
 *
 *<dt><b>History:</b>
 *<dd>		16Sep97	/Antonio Alexandre Rua	/New
 *<dd>		8Dec97	/Ronan Fauglas		/changes for documentation templates + complements in documentation
 *<dd>		11Dec97	/Ronan Fauglas		/changed access methods: thisVariable() --> getThisVariable()
 *
 *<dt><b>References:</b>
 *<dd>		DIS-Java-VRML Working Group: <a href="http://www.stl.nps.navy.mil/dis-java-vrml/"> 
 *		http://www.stl.nps.navy.mil/dis-java-vrml/</a>
 *<dd>		DIS specification : IEEE 1278.1, Section 5.3.29, 4.4.5.1
 *
 *@see ProtocolDataUnit
 *@see PduElement 
 *@see SerializationInterface
 */
public abstract class SimulationManagementPdu extends ProtocolDataUnit 
{
    /**
     *Identifier of the entity originating the PDU.
     *The semantics depend on derived class type.
     */
    protected EntityID      originatingEntityID;   // Originating Entity, semantics depend on derived class type

    /**
     *Identifier of the entity receiving the PDU.
     *The semantics depend on derived class type.
     */
    protected EntityID      receivingEntityID;     // Receiving Entity, semantics depend on derived class type


    /**
     *Constant value--size of a Simulation Management  PDU without headder.
     *<code>sizeOf = 12 bytes</code>
     */
    public static final int sizeOf = 12;
		// 12 bytes long (exclusive of ProtocolDataUnit general header data)
		// stands for originatingEntityID.sizeOf()+receivingEntityID.sizeOf()

  
public SimulationManagementPdu()						// default constructor
{
  originatingEntityID = new EntityID();
  receivingEntityID   = new EntityID();

  return;
}

public void  serialize(DataOutputStream outputStream)
{
 /**
    Serialize our data out to the stream. We first call the superclass,
    so that all the data there is written out to the buffer first.
    Then we serialize each of our ivars to the stream, being careful
    about the order.
*/
  // takes care of all the header variables in superclass
  super.serialize(outputStream);
  
  // write out the entity IDs...
  originatingEntityID.serialize(outputStream);
  receivingEntityID.serialize(outputStream);

  return;
}

public void  deSerialize(DataInputStream inputStream)
{
 /**
    deserialize our data from the stream. We first call the superclass,
    so that all the data there is read into the buffer first.
    Then we deserialize each of our ivars from the stream, being careful
    about the order.
 */
  //read protocol data unit general header first
  super.deSerialize(inputStream);

  // Read in Entity IDs...
  originatingEntityID.deSerialize(inputStream);
  receivingEntityID.deSerialize(inputStream);

  return;
}

public Object clone()
{
 /**
    Clone the SimulationManagementPdu, being careful to not share any pointers between the
    new object and the old object.
  */

	SimulationManagementPdu newPdu = (SimulationManagementPdu)super.clone();

	newPdu.setOriginatingEntityID(this.getOriginatingEntityID());
	newPdu.setReceivingEntityID(this.getReceivingEntityID());

	return newPdu;
}

public int length()
{
  return (super.length() + sizeOf);
}

/**
 *Returns the name of this PDU
 *
 *@return the name of this PDU
 */
public abstract String PDUName();


public void printValues(int indentLevel, PrintStream printStream)
{
  int superclassIndent = indentLevel;     // level of indentation for our superclass

  // ugly wart: get the superclass over to the left a couple pixels, if we have any to spare,
  // so the header info will be indented a bit less.
  if(superclassIndent > 0)
    superclassIndent -= 1;
  super.printValues(superclassIndent, printStream);

  originatingEntityID.printValues(indentLevel, printStream);
  receivingEntityID.printValues(indentLevel, printStream);

  return;
}


// Accessor methods

public EntityID getOriginatingEntityID()
{	return (EntityID)originatingEntityID.clone();
}
public void setOriginatingEntityID(EntityID pEntityID)
{	originatingEntityID = pEntityID;
}

public EntityID getReceivingEntityID()
{	return (EntityID)receivingEntityID.clone();
}
public void setReceivingEntityID(EntityID pEntityID)
{	receivingEntityID = pEntityID;
}

}   // end of class SimulationManagementPdu
