/*
 File:		SimulationManager.java
 CVS Info:	$Id: SimulationManager.java,v 1.1.1.1 1998/01/22 21:35:14 mcgredo Exp $
 Compiler:	jdk 1.1.6 
 */

package mil.navy.nps.dis;

import java.util.*;

import mil.navy.nps.util.*;
import mil.navy.nps.disEnumerations.*;
import mil.navy.nps.dis.*;
import mil.navy.nps.javaViaScriptNode.*;

/** 
 *@version 1.0
 *@author <a href="mailto:mcgredo@stl.nps.navy.mil">Don McGregor</a> (<a href="http://www.stl.nps.navy.mil/~mcgredo">http://www.stl.nps.navy.mil/~mcgredo</a>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/SimulationManager.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/SimulationManager.java</a>
 *
 *<dt><b>Hierarchy Diagram:</b>
 *<dd><a href="images/PduClassHierarchy.jpg"><IMG SRC="images/PduClassHierarchyButton.jpg" ALIGN=ABSCENTER></a>
 *
 *<dt><b>Summary:</b>
 *<dd> A single object, called SimulationManager, will handle all routing 
 * of PDUs. More functionality beyond this multiple-entity management
 * might be integrated into this object as well.
 *
 *<dt><b>Explanation</b>
 *<dd> The SimulationManager contains:
 * <ul>
 *     <li> A NetworkMonitor object
 *     <li> A list of all the active entities 
 * </ul>
 * <P>
 * Its responsibilities include
 * <ul>
 *   <li> Sending ESPDUs and any other entity-specific information to the 
 *    correct entity
 *   <li> Handling Simulation Management PDUs
 * </ul>
 * <P>
 * A block diagram of the object:
 * 
 * <PRE>
 *                   +---------------------------+
 *                   |                           |
 *                   |    SimulationManager      |
 *                   |                           |
 *                   |                           |
 *                   |              EntityList   |------VRML Scene
 *                   |                Entity1    |
 *                   |                Entity2    |
 *                   |                Entityn    |
 *                   |                           |
 *                   |           NIU             |
 *                   +---------------------------+
 *                                |
 *                       ----------------- Wire
 * </PRE>
 * The NIU, an instance of a NetworkMonitor, reads PDUs
 * from the wire. The SimulationManager occasionally pulls
 * PDUs from the NIU. For each PDU, the simulation manager
 * looks at the type. 
 * <P>
 * If it's an ESPDU, the SimulationManger code looks at the EntityID 
 * triplet (site, application, ID) that uniquely identifies each 
 * entity, then passes it on to the correct entity. If it's an
 * unrecognized EntityID, we may optionally instantiate a local
 * entity to reflect it, and insert it into the scene. The Entity
 * object in the list is responsible for communications with the
 * VRML scene. ESPDU culling may be performed either here or
 * at the entity level. It seems more logical to put it at
 * the entity level, where there is a bit more information about
 * update frequency and the like. On the other hand, it might
 * make more sense to cull ESPDUs early, to cut down on PDU 
 * handling overhead. That's an implementation and optimization detail.
 * <P>
 * If it's a SimulationManagement PDU, it will be handled by
 * a different block of code. SimulationManagement PDUs include
 * Start Simulation PDU, Stop Simulation PDU, Create Entity
 * PDU, EventReport PDU, Query PDU and Data PDU. Each of
 * these will probably require some special handling, and
 * perhaps some state changes in the SimulationManager object.
 * Some of them will require that PDUs be sent in response.
 * <P>
 * Cardinality: there is one and only one instance of a SimulationManager
 * per DIS application. I think the whole thing will operate
 * through class (static) methods.
 * <P>
 *
 *<dt><b>History:</b>
 *<dd>		29Oct97 /Don McGregor    	/New
 *<dd>		19May97  DonBrutzman	Added addEspduTransform
 *
 *@see NetworkMonitor
 *@see java.util.Hashtable
 *@see java.lang.Object
 *@see java.lang.Thread
 *@see java.lang.Runnable
 */

public class SimulationManager extends Object implements Runnable
{
  private static final boolean DEBUG = true;                          // Debugging/diagnostic output

  private        NetworkMonitor    niu=null;                          // Handles communication w/ the wire
  private        Hashtable         entities = new Hashtable();        // contains all ESPDUReadTransforms in the vrml world


/**
  Debugging output. Pass in a string, and it gets printed out on the console.
  You can pass in strings such as "foo " + "bar"
*/

protected static void debug(String pDiagnostic)
{
  if(DEBUG)
    System.out.println(pDiagnostic);
}

/**
    Unicast constructor; listens on the given unicast port
*/
public SimulationManager(int pPort)
{
  niu = new NetworkMonitor(pPort);

  debug("created unicast network monitor (this should appear only once)");
  return;
}

/**
 Multicast constructor; listens on the given port and group
*/

public SimulationManager(int pPort, String pMulticastGroup)
{
  niu = new NetworkMonitor(pMulticastGroup, pPort);
  
  debug("created multicast network monitor (this should appear only once)");

  return;
}

/**
 Adds a readTransform node to the list of readTransforms
 being maintained. Once a ReadTransform has been added to
 the list, it can receive PDUs sent to that EntityID triplet.
*/

public  void addReadTransform(EspduReadTransform pRT)
{
  EntityID  entID = pRT.getEntityIDObject();

  entities.put(entID, pRT);

  return;
}

public  void addEspduTransform(EspduTransform pET)
{
  EntityID  entID = pET.getEntityIDObject();

  entities.put(entID, pET);

  return;
}


public void run()   // loops endlessly
{
  Vector              newPdus;
  Enumeration         enumeration;
  ProtocolDataUnit    pdu;
  EspduReadTransform  espduRTNode = null;

  debug("Entering run loop");

  // Thread used to run the network monitor
  Thread aThread = new Thread(niu);

  debug("thread starting....");
  // Save the thread ID in the niu, then start the NIU reading.
  niu.setThread(aThread);
  aThread.start(); 

  debug("thread started....");

  while(true)
  {
    newPdus = niu.receivedPdus();

//**    debug("got " + newPdus.size() + " PDUs this cycle");

    enumeration = newPdus.elements();

    while(enumeration.hasMoreElements() )
    {
      pdu = (ProtocolDataUnit)enumeration.nextElement();

      switch((pdu.getPduType()).intValue())
      {
        // The major cases
        case  PDUTypeField.ENTITYSTATE:
                debug("Attempting to dispatch ESPDU to entity");

                // Find entity in database, dispatch pdu to that entity
                espduRTNode = (EspduReadTransform)(entities.get(((EntityStatePdu)pdu).getEntityID()));
                if(espduRTNode == null)
                {
                  EntityID  eid = ((EntityStatePdu)pdu).getEntityID();
                  debug("Couldn't find node that corresponds to EntityID");
                  debug("site: " + eid.getSiteID().intValue() + " application: " + eid.getApplicationID().intValue() +
                    "entityID: " + eid.getEntityID().intValue());
                }
                if(espduRTNode != null)
                  espduRTNode.addPDU(pdu);
              break;

              //case  PduTypeEnum.FIRE:
        //case  PduTypeEnum.DETONATION:

        
          // Some other type of PDU. If it's a simulation management PDU, that gets
          // handled in a method of its own.

        default:
              //if(ClassUtilities.objectIsSubclassOfClass(pdu, SimulationManagementPdu.getClass()))
              //    handleSimulationManagementPdu((SimulationManagementPdu)pdu);
              // else
                 System.out.println("Don't know how to handle PDU of type " + pdu.getPduType().intValue());

      } // End of switch
    }   // end of while enumeration
  }     // end of while true

  } // end of method

private void handleSimulationManagementPdu(SimulationManagementPdu pSimManagePdu)
{
  return;
}

} // end of class

      
