/*
 File:		EspduReadTransform.java
 CVS Info:	$Id: EspduReadTransform.java,v 1.3 1998/02/15 15:29:20 brutzman Exp $
 Compiler:	jdk 1.1.6 
 */

// **********************************************************************
//   IMPORT
// **********************************************************************

package mil.navy.nps.javaViaScriptNode;

import java.lang.*;
import java.util.*;
import java.io.*;

// Java JSAI classes for VRML

import vrml.*;
import vrml.field.*;
import vrml.node.*;

// NPS DIS library classes

import mil.navy.nps.dis.*;		
import mil.navy.nps.math.*;

/** 
 *@version 1.0
 *@author <a href="mailto:brutzman@nps.navy.mil">Don Brutzman</a> (<a href="http://www.stl.nps.navy.mil/~mcgredo">http://www.stl.nps.navy.mil/~brutzman</a>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/JavaViaScriptNode/EspduReadTransform.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/JavaViaScriptNode/EspduReadTransform.java</a>
 *
 *<dt><b>Summary:</b>
 *<dd> PROTO providing DIS ESPDU interface for a VRML entity
 *
 *<dt><b>Explanation</b>
 *<dd> This Java class provides the interface between the EspduReadTransform Script node
 * in the VRML scene, and the DIS class library.
 *
 *<dt><b>History:</b>
 *<dd>		1Aug97  /Don Brutzman    	/New
 *<dd>		5Feb98	/Don McGregor		/changes for multiple entities, using SimulationManager
 *<dd>		12Feb98	/Don Brutzman		/changes for documentation templates + complements in documentation
 *
 *<dt><b>Associated VRML Files:</b>
 *<dd>	<a href="EspduReadTransform.wrl">EspduReadTransform.wrl</A> (EXTERNPROTO)
 *<dd>	<a href="EspduReadTransformExample.wrl">EspduReadTransformExample.wrl</A> (Example Use)
 *
 *<dt><b>Unfinished business:</b>
 *<dd>		Need to verify PDU identity and match that given in scene
 *<dd>		Need to verify Euler angle order & quaternion conversion
 *<dd>		Is there another callback alternative to trigger processEvent originating from DIS rather than VRML?
 *<dd>		Is VRML timeSensor processEvent trigger best?  Likely so, since it ensures VRML scene redraw doesn't undergo starvation, and it also prevents "simultaneous" data access/modification of Transform node values by VRML browser/Java script.
 *<dd>		Exercise clock synchronization or relative local time? dead reckon interval dt is currently estimated by accumulating frame rate (redraw time) during each loop.
 *<dd>		Add algorithms for dead-reckoning and smoothing.
 *
 *@see EspduWriteTransform 
 */

// Notes:		System.out.println text appears on WorldView VRML console
//					(using Java via the Script node)
//
//				Incorporates parts of Kent Watsen's EAI-based World.java,
//				Ownship.java and Don McGregors's NetworkMonitorTest.java
//

public class EspduReadTransform extends Script 
{
/**
 * When DEBUG is true, System.out.println text messages trace the internals of script operation.
 * Text output appears in the Java Console (CosmoPlayer browser) or in the VRML console (WorldView browser)
 */
 
	public static boolean DEBUG		= true;  // diagnostic messages on/off

/**
 * 
 */
 
/**
 * Declare eventIns, fields and eventOuts from VRML scene
 */

/**
 * eventIn, trigger the read
 */
 
	private SFTime		readPDU;
/**
 * field, hook to VRML scene
 */
 
	private SFNode 		transformNode;
/**
 *  field, 11-character DIS marking
 */
 
	private SFString	marking;
/**
 * field, multicast address, or "unicast"
 */
 
	private SFString	address; 

/**
 * field, port to listen on
 */
 
	private SFInt32		port;

/**
 * field, how often to look for DIS PDUs
 */
 
	private SFTime		readInterval;

/**
 * field, entityID triplet: site ID
 */
 
  private SFInt32   siteID;

/**
 * field, entityID triplet: unique app ID at that site
 */
 
  private SFInt32   applicationID;

/**
 * field, entity ID triplet: ID within that app
 */
 
  private SFInt32   entityID; 

/**
 * Local variables
 */
 
	private Node		node;
	private SFVec3f		translation;		// get
	private SFRotation 	rotation;		// get
	private SFVec3f		position;		// set		
	private SFRotation 	orientation;		// set
	private Class		espduClass	= null;
	private EntityState	entity		= null;
	private	int 		pduCount	= 0;
	private Quaternion	quaternion	= null;
	private float		rot []		= null;
	private float		eulers []	= null;

  private EntityID    entityIDObject    = null;		// unique triplet that identifies entity

  private Vector      receivedPDUs = new Vector();	// holds PDUs that have come in

  public static SimulationManager simulationManager = null;  // routes PDUS to the VRML node corresponding to that entity

	private Browser		browser		= null; // VRML browser
	private float		dt		= 0.0f; // time step


/**
  Debugging output. Pass in a string, and it gets printed out on the console.
  You can pass in strings such as "foo " + "bar"
*/

protected static void debug(String pDiagnostic)
{
  if(DEBUG)
    System.out.println(pDiagnostic);
}

public String getName () { return "EspduReadTransform " + marking; }

/** For a good initialize () discussion, see example 3.15, pp. 95-97,
  * "Java for 3D and VRML Worlds," Lea/Matsuda/Miyashita, New Riders Press, 1996.
  */
  
public void initialize ()
	{
 		if (DEBUG)
 			System.out.println ("   EspduReadTransform:  initialize");

        try
        {
           Thread.sleep(500);          // go to sleep
        }
        catch(InterruptedException interruptedException)
        {
           throw new RuntimeException("EspduReadTransform:  exceptional sleep ");
        }
 			
	browser = getBrowser ();
	if (DEBUG) {
		System.out.println ("   EspduReadTransform:  browser.getName    () = "
		                                           + browser.getName ());
		System.out.println ("   EspduReadTransform:  browser.getVersion () = " 
		                                           + browser.getVersion ());
 		}
		// see examples 3.11/3.12 pp. 83-86
		// "Java for 3D and VRML Worlds," Lea/Matsuda/Miyashita

		transformNode = (SFNode) getField ("transformNode"); // instantiate
		node = (Node) (transformNode.getValue());	// access Transform node
		
		translation = (SFVec3f) node.getExposedField ("translation");	// get
		if (DEBUG) System.out.println
		           ("   initial transformNode.translation = " + translation);
		rotation = (SFRotation) node.getExposedField ("rotation");	// get
		if (DEBUG) System.out.println
		           ("   initial transformNode.rotation    = " + rotation);
	    try
		{
   			espduClass = Class.forName ("mil.navy.nps.dis.EntityStatePdu");
		}
    	catch (ClassNotFoundException e)
		{
			System.out.println
			               ("   EspduReadTransform:  EntityStatePdu class not found");
			System.out.println (e);
		}
    	if (DEBUG) System.out.println ("   EspduReadTransform:  espduClass found" );

		marking    = (SFString) getField ("marking");       // instantiate
		address    = (SFString) getField ("address");       // instantiate
		port       = (SFInt32)  getField ("port");          // instantiate
    siteID     = (SFInt32)  getField ("siteID");
    applicationID = (SFInt32)  getField ("applicationID");
    entityID   = (SFInt32)  getField ("entityID");      // instantiate 

    if (DEBUG) System.out.println
		           ("   EspduReadTransform:  marking         = " + marking);
		if (DEBUG) System.out.println
		           ("   EspduReadTransform:  network address = " + address);
		if (DEBUG) System.out.println
		           ("   EspduReadTransform:  port            = " + port);
    if (DEBUG) System.out.println
               ("   EspduReadTransform: siteID           = " + siteID);
    if (DEBUG) System.out.println
               ("   EspduReadTransform: applicationID    = " + applicationID);
    if (DEBUG) System.out.println
               ("   EspduReadTransform: entityID         = " + entityID);


    // Simulation manager is static; only one copy is shared between all instances
    // of the EspduReadTransform. If it's null the first time through, create it.

    if(simulationManager == null)
    {

      if (address.getValue().equalsIgnoreCase  ("unicast")) 
        simulationManager = new SimulationManager(port.getValue());   // unicast
      else
        simulationManager = new SimulationManager(port.getValue(), address.getValue()); // multicast

      // Thread used to run the SimulationManager
      Thread aThread = new Thread(simulationManager);
      aThread.start(); 
    }

    // DIS entityID object, which has a few more smarts.
    entityIDObject = new EntityID((short)siteID.getValue(), 
                                  (short)applicationID.getValue(), 
                                  (short)entityID.getValue());

    // Add ourselves to the list of entities that can receive PDUs.
    simulationManager.addReadTransform(this);

		readInterval   = (SFTime)  getField ("readInterval");   // instantiate
		if (true) System.out.println
		           ("   EspduReadTransform:  readInterval    = " + readInterval);
DEBUG=false;

		return;
  	}

/** Ignore all but latest event so only one set of PDU network reads
 * and one set of dead-reckon calculations occur.
 * This allows frequent updating triggered by the script eventIn.
 */

public void processEvents ( int count, Event events []) 
	{
		if (DEBUG) System.out.println
		           ("processEvents:  " + count + " events queued");

		processEvent (events [count-1]);
	}
	
/**
   addPDU() is called by the simulation manager when a PDU for the node
   arrives. The PDU is added to the list of PDUs that have already arrived,
   and is later processed when processEvent() is called.
*/

public void addPDU(ProtocolDataUnit pPDU)
{

  debug("Got a PDU in the ReadTransform node");

  receivedPDUs.addElement(pPDU);
  return;
}

/**
  gets the entity ID, which is the unique triplet that identifies the entity.
  The triplet is (site, application, ID). This is set in the VRML node, and
  retrieved for use here.
*/

public EntityID getEntityIDObject()
{
  return entityIDObject;
}

public void processEvent (Event event) // This is the main method
	{
		ProtocolDataUnit	pdu		= null;
		EntityID		entityID	= null;
		EntityStatePdu		espdu		= null;
		float			frameTime	= 0.0f;
		float			x, y, z, roll, pitch, yaw;

		quaternion  = new Quaternion ();
		orientation = new SFRotation ();
		eulers		= new float [3];
		rot 		= new float [4];

		if (DEBUG) System.out.println ("   EspduReadTransform:  processEvent ");

		if (DEBUG) System.out.println (event);


		int pduVectorSize = receivedPDUs.size ();
        	pduCount += pduVectorSize;
        
        if (pduVectorSize <= 0)
        {
 		   if (DEBUG) System.out.println ("Received 0 PDUs [" + pduCount + " total]");
		   return;
        }
       	else if (DEBUG) System.out.println
		    ("Received " + pduVectorSize + " PDUs [" + pduCount + " total]");

		// estimate dead-reckon interval dt, zeroed later if new PDU has arrived
		if (readInterval.getValue () < 1.0)
		{
			 frameTime	= 1.0f / browser.getCurrentFrameRate ();
			 dt 		= dt + frameTime;
		}
	 // else long time between PDU reads
		else dt = dt + (float) readInterval.getValue (); 
		
 		if (DEBUG) {
 		//	System.out.println ("   EspduReadTransform:  browser.getCurrentSpeed () = "
 		//	                                           + browser.getCurrentSpeed ());
 			System.out.println ("   EspduReadTransform:  browser.getCurrentFrameRate () = "
 			                                           + browser.getCurrentFrameRate ());
 			System.out.println ("   EspduReadTransform:  dt = " + dt);
 		}

		// process all received ESPDUs, update EntityState hash table for all
		//   entities received (possibly inefficient), drop (!) other PDU types
		    
		for (int i = 0; i < pduVectorSize; i++) 
        {
        	pdu = (ProtocolDataUnit) receivedPDUs.elementAt (i);
        	if (pdu.getClass() == espduClass) // verify this is an Entity State PDU
          	{
          		espdu    = (EntityStatePdu)pdu;
          		entityID = espdu.getEntityID();

          		entity = EntityState.get(entityID);  // check if entity already exists
          		if (entity == null) // first time seeing this entity
            	{
            		entity = new EntityState (entityID);
            	}
            	// !!! to do:  need to verify correct identity
            	
				entity.setESPDU(espdu);	// go ahead and set pdu data now
				dt = 0.0f;				// reset local dead-reckon clock
			}
        }
		receivedPDUs.removeAllElements(); // garbage collector will destroy Vector

		// let each entity update itself
		for (int i=EntityState.numberEntities (); i > 0; i--)
		    // backwards 'cause delete shifts list
        {
        	entity = EntityState.get(i-1);

			if (entity.getPduTime() > 10.0) // 10 sec timeout, make this selectable
			{
				if (DEBUG) System.out.println ("entity 10 sec timeout:  " + entity);
				entity.dispose();
			}
		}
		node = (Node) (transformNode.getValue()); // get VRML transformNode

		// verify proper espdu in hand??
        // ... awaiting multiple entity modifications by Don McGregor ...

		if (DEBUG) System.out.println
           ("EspuReadTransform.java: attempting to read marking from espdu");
          
        marking.setValue (espdu.getMarking ());
        
        if (DEBUG) System.out.println
           ("EspuReadTransform.java: made it past read of marking from espdu");

		// Entity coordinate systems (right-hand rule applies)

		//          DIS                 VRML
		//
		//         -Z entity up         Y entity up
		//          |                   |
		//          |                   |
		//          |                   |
		//          |                   |
		//          +------- X          +-------- X    nose of entity body
		//         /                   /
		//        /                   /
		//       /                   /
		//      Y                   Z                   
		//         right-hand side
		//         of entity body
		//
		// Rotation angle axes (right-hand rule applies):
		//
		//          DIS                 VRML          Rotation angle synonyms
		//
		// Roll     X                    X            phi
		// Pitch    Y                    Z            theta
		// Yaw		Z					-Y            psi

		// Calculate VRML position/orientation from DIS position/orientation:
		
	 	x = (float)   espdu.getEntityLocationX()  + dt * espdu.getEntityLinearVelocityX();
	 	y = (float) (-espdu.getEntityLocationZ()) - dt * espdu.getEntityLinearVelocityZ();
	 	z = (float)   espdu.getEntityLocationY()  + dt * espdu.getEntityLinearVelocityY();

	 	position    = new SFVec3f (	x, y, z );
		if (DEBUG) System.out.println("position = (" + x + ", " + y + ", " + z + ")");

		roll = espdu.getEntityOrientationPhi()   + dt * espdu.getEntityAngularVelocityX();
		pitch= espdu.getEntityOrientationTheta() + dt * espdu.getEntityAngularVelocityY();
		yaw  = espdu.getEntityOrientationPsi()   + dt * espdu.getEntityAngularVelocityZ();

		if (true) System.out.println ("phi/theta/psi = "
				+ roll  * 180.0 / Math.PI + ", "
				+ pitch * 180.0 / Math.PI + ", "
				+ yaw   * 180.0 / Math.PI + " degrees");

		// note Kent's quaternion code has irregular ordering of Euler angles
		// which (by whatever method :) accomplishes the angle transformation
		// desired...   (verified using NPS AUV)
				
	 	eulers [0] = -yaw;
	 	eulers [1] =  roll;
		eulers [2] =  pitch;
		
		quaternion.setEulers (eulers);

	 	quaternion.getAxisAngle (rot);
	 	orientation.setValue 	(rot);
		if (DEBUG) System.out.println("rotation values = (" + rot[0] + ", " +
		              rot[1] + ", " + rot[2] + ", " + rot[3] + ")" );
	 	
		translation.setValue (position);								// set
		if (DEBUG) System.out.println("transformNode.translation = " + position);
	 	
		rotation.setValue (orientation);								// set
		if (DEBUG) System.out.println("TransformNode.rotation = " + orientation);

		return;
	}
}


