/*
 File:		EspduReadTransform.java
 CVS Info:	$Id: EspduReadTransform.java,v 1.3 1998/02/15 15:29:20 brutzman Exp $
 Compiler:	jdk 1.1.6 
 */

// **********************************************************************
//   IMPORT
// **********************************************************************

package mil.navy.nps.javaViaScriptNode;

import java.lang.*;
import java.util.*;
import java.io.*;

// Java via Script Authoring Interface (JSAI) classes for VRML

import vrml.*;
import vrml.field.*;
import vrml.node.*;

// NPS DIS library classes

import mil.navy.nps.dis.*;		
import mil.navy.nps.math.*;

/** 
 *@version 1.0
 *@author <a href="mailto:brutzman@nps.navy.mil">Don Brutzman</a> (<a href="http://www.stl.nps.navy.mil/~mcgredo">http://www.stl.nps.navy.mil/~brutzman</a>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/JavaViaScriptNode/EspduReadTransform.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/JavaViaScriptNode/EspduReadTransform.java</a>
 *
 *<dt><b>Summary:</b>
 *<dd> PROTO providing DIS ESPDU read/write interface for a VRML entity.
 *
 *<dt><b>Explanation</b>
 *<dd> This Java class provides the interface between the EspduReadTransform Script node
 * in the VRML scene, and the DIS class library.
 *
 *<dt><b>History:</b>
 *<dd>		19May98  /Don Brutzman    	/New, combines previous read & write versions into a single class.
 *
 *<dt><b>Associated VRML Files:</b>
 *<dd>	<a href="../mil/navy/nps/javaViaScriptNode/EspduTransformPROTO.wrl">EspduTransformPROTO.wrl</A> (EXTERNPROTO)
 *<dd>	<a href="../mil/navy/nps/javaViaScriptNode/EspduTransformExample.wrl">EspduTransformExample.wrl</A> (example use)
 *
 *<dt><b>Debugging:</b>
 *<dd>	DEBUG=false turns off debug/trace messages
 *<dd>	INITIALIZE_DEBUG=true turns on debug/trace messages during initialize ()
 *<dd>	System.out.println text output appears either on the browswer's Java console or VRML console
 *
 *<dt><b>Unfinished business:</b>
 *<dd>		Need to verify PDU identity and match that given in scene
 *<dd>		Need to verify Euler angle order & quaternion conversion
 *<dd>		Is there another callback alternative to trigger processEvent originating from DIS rather than VRML?
 *<dd>		Is VRML timeSensor processEvent trigger best?  Likely so, since it ensures VRML scene redraw doesn't undergo starvation, and it also prevents "simultaneous" data access/modification of Transform node values by VRML browser/Java script.
 *<dd>		Exercise clock synchronization or relative local time? dead reckon interval dt is currently estimated by accumulating frame rate (redraw time) during each loop.
 *<dd>		Add algorithms for dead-reckoning and smoothing.
 *<dd>		Ensure all exposedField parameters can be reset during operation.
 *
 *@see mil.navy.nps.dis.SimulationManager
 *@see mil.navy.nps.dis.EntityStatePdu
 */

//				Incorporates parts of Kent Watsen's EAI-based World.java,
//				Ownship.java and Don McGregors's NetworkMonitorTest.java

public class EspduTransform extends Script 
{
/**
 * When DEBUG is true, System.out.println text messages trace the internals of script operation.
 * Text output appears in the Java Console (CosmoPlayer browser) or in the VRML console (WorldView browser)
 */
 
	public static boolean DEBUG		= false;  // diagnostic messages on/off
	public static boolean INITIALIZE_DEBUG	= true;   // diagnostic messages on/off

/**
 * Declare eventIns, fields and eventOuts unique to the EspduTransform PROTO
 */

/**
 * eventIn:  trigger the read
 */
 
	private SFTime		readPDU;
/**
 * field:  hook to VRML scene
 */
 
	private SFNode 		transformNode;

/**
 * Declare eventIns, fields and eventOuts from the user's VRML scene
 */

/**
 * field:  a multicast address, or the literal string "unicast"
 */
 
	private SFString	address; 

/**
 * field:  port to listen on
 */
 
	private SFInt32		port;

/**
 * field:  how often to query NetworkMonitor to read/write DIS PDUs
 */
 
	private SFTime		period;

/**
 *  field:  11-character DIS marking
 */
 
	private SFString	marking;
/**
 * field:  entityID triplet: site ID
 */
 
  private SFInt32   siteID;

/**
 * field:  entityID triplet: unique app ID at that site
 */
 
  private SFInt32   applicationID;

/**
 * field:  entity ID triplet: ID within that app
 */
 
  private SFInt32   entityID; 

/**
 * Local variables
 */
 
	private Node		node;
	private SFVec3f		translation;		// get
	private SFRotation 	rotation;		// get
	private SFVec3f		position;		// set		
	private SFRotation 	orientation;		// set
	private Class		espduClass	= null;
	private EntityState	entity		= null;
	private	int 		pduCount	= 0;
	private Quaternion	quaternion	= null;
	private float		rot []		= null;
	private float		eulers []	= null;

	static final float	initPos[]		 = {0f, 5f, 0f}; // x, y, z
	static final float	initPosDt[]	= {0f, 0f, 0f}; // aka: velocity
	static final float	initAng[]		 = {0f, 0f, 0f}; // heading, pitch, roll
	static final float	initAngDt[]	= {0f, 0f, 0f}; // aka: angular rates

	Quaternion		tmpQuat		= null;
	
	float			pos[]		= null; // x, y, z
	float			posDt[]		= null; // aka: velocity
	float			ang[]		= null; // heading, pitch, roll
	float			angDt[]		= null; // aka: angular rates
	float			drPos[]		= null; // x, y, z
	float			drPosDt[]	= null; // aka: velocity
	float			drAng[]		= null; // heading, pitch, roll
	float			drAngDt[]	= null; // aka: angular rates

	private EntityID	entityIDObject	= null;		// unique triplet that identifies entity

	private Vector		receivedPDUs	= new Vector();	// holds PDUs that have come in

	public static SimulationManager simulationManager = null;  // routes PDUS to the VRML node corresponding to that entity

	private Browser		browser		= null; // VRML browser
	private float		dt		= 0.0f; // time step


/**
  Debugging output. Pass in a string, and it gets printed out on the console.
  You can pass in strings such as "foo " + "bar"
*/

protected static void debug(String pDiagnostic)
{
  if(DEBUG)
    System.out.println(pDiagnostic);
}

public String getName () { return "EspduTransform " + marking; }

/** For a good initialize () discussion, see example 3.15, pp. 95-97,
  * "Java for 3D and VRML Worlds," Lea/Matsuda/Miyashita, New Riders Press, 1996.
  */
  
public void initialize ()
{
 	if (INITIALIZE_DEBUG) System.out.println ("   EspduTransform:  initialize");

        try
        {
           Thread.sleep(500);          // go to sleep, msec
        }
        catch(InterruptedException interruptedException)
        {
           throw new RuntimeException("EspduTransform:  exceptional sleep ");
        }
 			
	browser = getBrowser ();
	if (INITIALIZE_DEBUG) {
		System.out.println ("   EspduTransform:  browser.getName    () = "
		                                       + browser.getName    ());
		System.out.println ("   EspduTransform:  browser.getVersion () = " 
		                                       + browser.getVersion ());
 	}
	// see examples 3.11/3.12 pp. 83-86
	// "Java for 3D and VRML Worlds," Lea/Matsuda/Miyashita

	// instantiations
	address      = (SFString) getField ("address");
	port         = (SFInt32)  getField ("port");
	period       = (SFTime)   getField ("period");
	marking      = (SFString) getField ("marking");
	siteID       = (SFInt32)  getField ("siteID");
	applicationID= (SFInt32)  getField ("applicationID");
	entityID     = (SFInt32)  getField ("entityID");
		
	if (INITIALIZE_DEBUG) {
	System.out.println ("   EspduTransform:  network address = " + address);
	System.out.println ("   EspduTransform:  port            = " + port);
	System.out.println ("   EspduTransform:  period          = " + period);
	System.out.println ("   EspduTransform:  marking         = " + marking);
	System.out.println ("   EspduTransform:  siteID          = " + siteID);
	System.out.println ("   EspduTransform:  applicationID   = " + applicationID);
	System.out.println ("   EspduTransform:  entityID        = " + entityID);
	}

	transformNode = (SFNode) getField ("transformNode"); // instantiate
	node = (Node) (transformNode.getValue());	     // access Transform node	
	translation = (SFVec3f) node.getExposedField ("translation");	// get
	rotation = (SFRotation) node.getExposedField ("rotation");	// get
	if (DEBUG)
	{
	  System.out.println ("   initial transformNode.translation = " + translation);
	  System.out.println ("   initial transformNode.rotation    = " + rotation);
	}

	try
	{
   		espduClass = Class.forName ("mil.navy.nps.dis.EntityStatePdu");
	}
    	catch (ClassNotFoundException e)
	{
		System.out.println
	               ("   EspduTransform:  EntityStatePdu class not found");
		System.out.println (e);
	}
    	if (INITIALIZE_DEBUG) System.out.println ("   EspduTransform:  espduClass found" );

	// Simulation manager is static; only one copy is shared between all instances
	// of the EspduTransform.  If null, this is the first time through, create it.
	if (simulationManager == null)
	{
 	  if (INITIALIZE_DEBUG) System.out.println ("   EspduTransform:  starting SimulationManager");
	  if (address.getValue().equalsIgnoreCase  ("unicast")) 
		simulationManager = new SimulationManager(port.getValue());   // unicast
	  else
		simulationManager = new SimulationManager(port.getValue(), address.getValue()); // multicast

	  // Thread used to run the SimulationManager
 	  if (INITIALIZE_DEBUG) System.out.println ("   EspduTransform:  starting new thread");
	  Thread aThread = new Thread(simulationManager);
	  aThread.start();
 	  if (INITIALIZE_DEBUG) System.out.println ("   EspduTransform:  aThread.start() complete");
	}

	// DIS entityID object, which has a few more smarts & must know identity triplet.
	entityIDObject = new EntityID(	(short)siteID.getValue(), 
					(short)applicationID.getValue(), 
					(short)entityID.getValue());

	// high-fidelity variables (local best representation)
	pos	= new float[3]; // x, y, z
	posDt	= new float[3]; // aka: velocity
	ang	= new float[3]; // heading, pitch, roll
	angDt	= new float[3]; // aka: angular rates

	// low-fidelity variables
	drPos	= new float[3]; // x, y, z
	drPosDt	= new float[3]; // aka: velocity
	drAng	= new float[3]; // heading, pitch, roll
	drAngDt	= new float[3]; // aka: angular rates

	// these are used for hpr->axis/angle conversions
	tmpQuat	= new Quaternion();
	rot	= new float[4]; // axis/angle

	// now set init vars
	System.arraycopy(initPos,   0, pos,   0, 3);
	System.arraycopy(initPosDt, 0, posDt, 0, 3);
	System.arraycopy(initAng,   0, ang,   0, 3);
	System.arraycopy(initAngDt, 0, angDt, 0, 3);

	// now convert heading-pitch-roll hpr (ang) to axis/angle (rot)
	tmpQuat.setEulers(ang);
	tmpQuat.getAxisAngle(rot);

	// Add ourselves to the list of entities that can receive PDUs.
	simulationManager.addEspduTransform(this);
 	if (INITIALIZE_DEBUG) System.out.println ("   EspduTransform:  added to simulationManager");


 	if (INITIALIZE_DEBUG) System.out.println ("   EspduTransform:  initialize () complete.");

	return;
}

/** Ignore all but latest event so only one set of PDU network reads
 * and one set of dead-reckon calculations occur.
 * This allows frequent updating triggered by the script eventIn.
 * This might need to be made smarter someday.
 */

public void processEvents ( int count, Event events []) 
{
	if (DEBUG) System.out.println ("processEvents:  " + count + " events queued, flush earliest");

	processEvent (events [count-1]);
}
	
/**
   addPDU() is called by the simulation manager when a PDU for the node
   arrives. The PDU is added to the list of PDUs that have already arrived,
   and is later processed when processEvent() is called.
*/

public void addPDU(ProtocolDataUnit pPDU)
{

	debug("Got a PDU in the ReadTransform node");

	receivedPDUs.addElement(pPDU);
	return;
}

/**
  gets the entity ID, which is the unique triplet that identifies the entity.
  The triplet is (site, application, ID). This is set in the VRML node, and
  retrieved for use here.
*/

public EntityID getEntityIDObject()
{
	return entityIDObject;
}

public void processEvent (Event event) // This is the main method
{
	ProtocolDataUnit	pdu		= null;
	EntityID		entityID	= null;
	EntityStatePdu		espdu		= null;
	float			frameTime	= 0.0f;
	float			x, y, z, roll, pitch, yaw;

	quaternion  = new Quaternion ();
	orientation = new SFRotation ();
	eulers		= new float [3];
	rot 		= new float [4];

	if (DEBUG) System.out.println ("   EspduTransform:  processEvent ");

	if (DEBUG) System.out.println (event);

	int pduVectorSize = receivedPDUs.size ();
	pduCount += pduVectorSize;
        
	if (pduVectorSize <= 0)
	{
 		   if (DEBUG) System.out.println ("Received 0 PDUs [" + pduCount + " total]");
		   return;
	}
       	else if (DEBUG) System.out.println
		    ("Received " + pduVectorSize + " PDUs [" + pduCount + " total]");

		// estimate dead-reckon interval dt, zeroed later if new PDU has arrived
		if (period.getValue () < 1.0)
		{
			 frameTime	= 1.0f / browser.getCurrentFrameRate ();
			 dt 		= dt + frameTime;
		}
	 // else long time between PDU reads
	else dt = dt + (float) period.getValue (); 
		
	if (DEBUG) {
	//	System.out.println ("   EspduTransform:  browser.getCurrentSpeed () = "
	//					       + browser.getCurrentSpeed ());
		System.out.println ("   EspduTransform:  browser.getCurrentFrameRate () = "
						       + browser.getCurrentFrameRate ());
		System.out.println ("   EspduTransform:  dt = " + dt);
 	}

	// process all received ESPDUs, update EntityState hash table for all
	//   entities received (possibly inefficient), drop (!) other PDU types
		    
	for (int i = 0; i < pduVectorSize; i++) 
	{
		pdu = (ProtocolDataUnit) receivedPDUs.elementAt (i);
		if (pdu.getClass() == espduClass) // verify this is an Entity State PDU
	  	{
	  		espdu    = (EntityStatePdu)pdu;
	  		entityID = espdu.getEntityID();

	  		entity = EntityState.get(entityID);  // check if entity already exists
	  		if (entity == null) // first time seeing this entity
			{
	    			entity = new EntityState (entityID);
			}
	    		// !!! to do:  need to verify correct identity
	    	
			entity.setESPDU(espdu);	// go ahead and set pdu data now
			dt = 0.0f;				// reset local dead-reckon clock
		}
	}
	receivedPDUs.removeAllElements(); // garbage collector will destroy Vector

	// let each entity update itself
	for (int i=EntityState.numberEntities (); i > 0; i--)
	// backwards because delete shifts list
	{
		entity = EntityState.get(i-1);

		if (entity.getPduTime() > 10.0) // 10 sec timeout, make this selectable
		{
			if (DEBUG) System.out.println ("entity 10 sec timeout:  " + entity);
			entity.dispose();
		}
	}
	node = (Node) (transformNode.getValue()); // get VRML transformNode

	// verify proper espdu in hand??
	// ... awaiting multiple entity modifications by Don McGregor ...

	if (DEBUG) System.out.println
	   ("EspuReadTransform.java: attempting to read marking from espdu");
	  
	marking.setValue (espdu.getMarking ());
	
	if (DEBUG) System.out.println
	   ("EspuReadTransform.java: made it past read of marking from espdu");

		// Entity coordinate systems (right-hand rule applies)

		//	  DIS		 VRML
		//
		//	 -Z entity up	 Y entity up
		//	  |		   |
		//	  |		   |
		//	  |		   |
		//	  |		   |
		//	  +------- X	  +-------- X    nose of entity body
		//	 /		   /
		//	/		   /
		//       /		   /
		//      Y		   Z		   
		//	 right-hand side
		//	 of entity body
		//
		// Rotation angle axes (right-hand rule applies):
		//
		//	  DIS		 VRML	  Rotation angle synonyms
		//
		// Roll     X		    X	    phi
		// Pitch    Y		    Z	    theta
		// Yaw		Z					-Y	    psi

		// Calculate VRML position/orientation from DIS position/orientation:
		
	 	x = (float)   espdu.getEntityLocationX()  + dt * espdu.getEntityLinearVelocityX();
	 	y = (float) (-espdu.getEntityLocationZ()) - dt * espdu.getEntityLinearVelocityZ();
	 	z = (float)   espdu.getEntityLocationY()  + dt * espdu.getEntityLinearVelocityY();

	 	position    = new SFVec3f (	x, y, z );
		if (DEBUG) System.out.println("position = (" + x + ", " + y + ", " + z + ")");

		roll = espdu.getEntityOrientationPhi()   + dt * espdu.getEntityAngularVelocityX();
		pitch= espdu.getEntityOrientationTheta() + dt * espdu.getEntityAngularVelocityY();
		yaw  = espdu.getEntityOrientationPsi()   + dt * espdu.getEntityAngularVelocityZ();

		if (true) System.out.println ("phi/theta/psi = "
				+ roll  * 180.0 / Math.PI + ", "
				+ pitch * 180.0 / Math.PI + ", "
				+ yaw   * 180.0 / Math.PI + " degrees");

		// note Kent's quaternion code has irregular ordering of Euler angles
		// which (by whatever method :) accomplishes the angle transformation
		// desired...   (verified using NPS AUV)
				
	 	eulers [0] = -yaw;
	 	eulers [1] =  roll;
		eulers [2] =  pitch;
		
		quaternion.setEulers (eulers);

	 	quaternion.getAxisAngle (rot);
	 	orientation.setValue 	(rot);
		if (DEBUG) System.out.println("rotation values = (" + rot[0] + ", " +
			      rot[1] + ", " + rot[2] + ", " + rot[3] + ")" );
	 	
		translation.setValue (position);								// set
		if (DEBUG) System.out.println("transformNode.translation = " + position);
	 	
		rotation.setValue (orientation);								// set
		if (DEBUG) System.out.println("TransformNode.rotation = " + orientation);

		return;
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

	public void processWriteEvent (Event event) 
	{
		EntityStatePdu		espdu	= new EntityStatePdu ();
		EntityID			entityID	= null;
		float				frameTime	= 0.0f;
		float				x, y, z, roll, pitch, yaw;

        SFVec3f 			position    = new SFVec3f ();

		orientation = new SFRotation ();
		eulers		= new float [3];
		rot 		= new float [4];

		if (DEBUG) System.out.println
				("  EspduWriteTransform:  processEvent");

		frameTime	= 1.0f / browser.getCurrentFrameRate ();
		dt 			= dt + frameTime;

		if (DEBUG) {
 		//	System.out.println ("  EspduWriteTransform:  browser.getCurrentSpeed () = "
 		//	                                           + browser.getCurrentSpeed ());
 			System.out.println ("  EspduWriteTransform:  browser.getCurrentFrameRate () = "
 			                                           + browser.getCurrentFrameRate ());
 			System.out.println ("  EspduWriteTransform:  cumulative dt = " + dt);
 		}

		pduCount ++;

		node = (Node) (transformNode.getValue()); // get VRML transformNode

 		if (DEBUG) System.out.println ("got transformNode.getValue() OK...");

		// verify proper espdu in hand??
        // ... awaiting multiple entity modifications by Don McGregor ...
		
 		if (DEBUG) System.out.println ("espdu name   = " + espdu.PDUName ());

// this fails, reason still unknown..
//		espdu.setMarking (((SFString) node.getExposedField ("marking")).getValue ());
//		if (DEBUG) System.out.println ("finished espdu.setMarking (..marking..) ");


		// Entity coordinate systems (right-hand rule applies)

		//          DIS                 VRML
		//
		//         -Z entity up         Y entity up
		//          |                   |
		//          |                   |
		//          |                   |
		//          |                   |
		//          +------- X          +-------- X    nose of entity body
		//         /                   /
		//        /                   /
		//       /                   /
		//      Y                   Z                   
		//         right-hand side
		//         of entity body
		//
		// Rotation angle axes (right-hand rule applies):
		//
		//          DIS                 VRML          Rotation angle synonyms
		//
		// Roll     X                    X            phi
		// Pitch    Y                    Z            theta
		// Yaw		Z					-Y            psi

		// Calculate VRML position/orientation from DIS position/orientation:
		
		position = (SFVec3f) node.getExposedField ("translation");	// get
		if (DEBUG) System.out.println("transformNode.translation = " + position);

        x = position.getX ();
        y = position.getY ();
        z = position.getZ ();

        espdu.setEntityLocationX (x);
        espdu.setEntityLocationY (y);
        espdu.setEntityLocationZ (z);

		if (DEBUG) System.out.println("position = (" + x + ", " + y + ", " + z + ")");

        espdu.setEntityLinearVelocityX (0.0f);
        espdu.setEntityLinearVelocityY (0.0f);
        espdu.setEntityLinearVelocityZ (0.0f);

		orientation = (SFRotation) node.getExposedField ("rotation");	// get
		if (DEBUG) System.out.println("transformNode.rotation = " + orientation);

		orientation.getValue (rot);

		if (DEBUG) System.out.println("rotation values = (" + rot[0] + ", " +
		                               rot[1] + ", " + rot[2] + ", " + rot[3] + ")" );

		quaternion.setAxisAngle (rot);
		
		if (DEBUG) { System.out.print ("quaternion " );  quaternion.print (); }

		quaternion.getEulers (eulers);

		// note Kent's quaternion code has irregular ordering of Euler angles
		// which (by whatever method :) accomplishes the angle transformation
		// desired...   (needs to be verified using NPS AUV)
				
		yaw   = -eulers [0];
		roll  =  eulers [1];
		pitch =  eulers [2];

		// set roll, pitch, and yaw in the PDU
		
        espdu.setEntityOrientationPsi (roll);
        espdu.setEntityOrientationTheta (pitch);
        espdu.setEntityOrientationPhi (yaw);

		if (DEBUG) System.out.println ("phi/theta/psi = "
				+ roll  * 180.0 / Math.PI + ", "
				+ pitch * 180.0 / Math.PI + ", "
				+ yaw   * 180.0 / Math.PI + " degrees");

		// This method is called at writeInterval durations
		// no dead reckoning needed, just use values in the VRML scene
		// we are here in the processEvent method so it is OK to send the PDU

		espdu.printValues (2, System.out);  // (indentLevel, printStream)

/// ???        networkMonitor.sendPDU (espdu, address.getValue(), port.getValue());
		
		return;
	}

}

