// **********************************************************************
// 
// Script Name:	EspduWriteTransform.java
//
// Description:	PROTO providing DIS ESPDU interface for a VRML entity
//
// Author:		Don Brutzman
// Revised:		4 September 97
//
// VRML:		EspduWriteTransform.wrl and EspduWriteTransformExample.wrl
//
// URL:	http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/JavaViaScriptNode/EspduWriteTransform.java
//
// Notes:		System.out.println text appears on WorldView VRML console
//					(using Java via the Script node)
//
//				Incorporates parts of Kent Watsen's EAI-based World.java,
//				Ownship.java and Don McGregors's NetworkMonitorTest.java
//
//				Multicast use still incompatible with Netscape 3.01 so not offered
//
//				Need to verify Euler angle order & quaternion conversion
//
//				Is VRML timeSensor processEvent trigger best?  Likely so,
//					since it tells when to send explicitly.  Future work:
//					send based on dead-reckon requirements with frequent
//					recall, insert velocity values since last PDU
//
//				Exercise clock synchronization or relative local time?
//					clock value for PDU not yet set properly.
//
// **********************************************************************

// **********************************************************************
//   IMPORT
// **********************************************************************

package mil.navy.nps.javaViaScriptNode;

import java.net.*;
import java.util.*;

// Java JSAI classes for VRML
import vrml.*;
import vrml.field.*;
import vrml.node.*;

// NPS DIS library classes
import mil.navy.nps.dis.*;		
import mil.navy.nps.math.*;

// **********************************************************************
//   CLASS
// **********************************************************************

public class EspduWriteTransform extends Script 
{
	public static boolean DEBUG				= true;  // diagnostic

	// Declare eventIns, fields and eventOuts from VRML scene
	
	private SFTime		writePDU;			// eventIn
	private SFNode 		transformNode;		// field

	private SFString	marking;			// field
	private SFString	address;			// field
	private SFInt32		port;				// field
	private SFTime		writeInterval;		// field
	
	// local variables
	private Node		node;

	private SFVec3f		translation;		// field
	private SFRotation 	rotation;			// field
	private SFVec3f		position;			// set		
	private SFRotation 	orientation;		// set

	private Class		espduClass      	= null;
	private EntityState	entity          	= null;
	private	int 		pduCount 			= 0;

	private Quaternion	quaternion			= new Quaternion ();

	private float		rot []				= null;
	private float		eulers []			= null;

	public static NetworkMonitor  networkMonitor;
	public static Thread          aThread;

	private Browser		browser				= null;
	private float		dt					= 0.0f;

  static final float  initPos[]     = {0f, 5f, 0f}; // x, y, z
  static final float  initPosDt[]   = {0f, 0f, 0f}; // aka: velocity
  static final float  initAng[]     = {0f, 0f, 0f}; // heading, pitch, roll
  static final float  initAngDt[]   = {0f, 0f, 0f}; // aka: angular rates

  Quaternion          tmpQuat       = null;
	
  float               pos[]         = null; // x, y, z
  float               posDt[]       = null; // aka: velocity
  float               ang[]         = null; // heading, pitch, roll
  float               angDt[]       = null; // aka: angular rates
  float               drPos[]       = null; // x, y, z
  float               drPosDt[]     = null; // aka: velocity
  float               drAng[]       = null; // heading, pitch, roll
  float               drAngDt[]     = null; // aka: angular rates

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

	public String getName () { return "EspduWriteTransform " + marking; }

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

	public void initialize ()
	{
		// see example 3.15, pp. 95-97
		// "Java for 3D and VRML Worlds," Lea/Matsuda/Miyashita

		InetAddress  inetAddress = null;

		browser = getBrowser ();
 		if (DEBUG) {
 			System.out.println ("  EspduWriteTransform:  browser.getName () = "
 			                                           + browser.getName ());
 			System.out.println ("  EspduWriteTransform:  browser.getVersion () = " 
 			                                           + browser.getVersion ());
 		}
		// see examples 3.11/3.12 pp. 83-86
		// "Java for 3D and VRML Worlds," Lea/Matsuda/Miyashita

		transformNode = (SFNode) getField ("transformNode"); // instantiate
		node = (Node) (transformNode.getValue());	// access Transform node
		
		translation = (SFVec3f) node.getExposedField ("translation");	// get
		if (DEBUG) System.out.println
		           ("   initial transformNode.translation = " + translation);
		rotation = (SFRotation) node.getExposedField ("rotation");	// get
		if (DEBUG) System.out.println
		           ("   initial transformNode.rotation    = " + rotation);
	    try
		{
   			espduClass = Class.forName ("mil.navy.nps.dis.EntityStatePdu");
		}
    	catch (ClassNotFoundException e)
		{
			System.out.println
			               ("  EspduWriteTransform:  EntityStatePdu class not found");
			System.out.println (e);
		}
    	if (DEBUG) System.out.println ("   EspduWriteTransform:  espduClass found" );

		marking    = (SFString) getField ("marking");       // instantiate
		address    = (SFString) getField ("address");       // instantiate
		port       = (SFInt32)  getField ("port");          // instantiate
		if (DEBUG) System.out.println
		           ("   EspduWriteTransform:  marking         = " + marking );
		if (DEBUG) System.out.println
		           ("   EspduWriteTransform:  network address = " + address + 
 			                                          ", port = " + port);
		           
		writeInterval   = (SFTime)  getField ("writeInterval");   // instantiate
		if (true) System.out.println
		           ("   EspduWriteTransform:  writeInterval   = " + writeInterval);
		           
        // OK to run NetworkMonitor on same address/port multiple times?
		// multicast appears reusable (standalone test by defaultBridge, incomplete)

/* inetAddress.isMulticastAddress() will not work until browser goes to JDK 1.1  :(
   therefore everything must be unicast until then
        try
        {
		 inetAddress = InetAddress.getByName (address.getValue());
		 System.out.println("  EspduWriteTransform: Created internet address");
         System.out.println("  EspduWriteTransform: multicast boolean is " + inetAddress.isMulticastAddress());
		 }
		catch(UnknownHostException e)
	 	{
	 		System.out.println("cannot resolve address for destination " + address.getValue() + ", constructor not possible");
	 		return;
	 	}

    	if (!(inetAddress.isMulticastAddress()))
*/
    	if (true) // force unicast
    	{
			if (DEBUG) System.out.println
			("   EspduWriteTransform:  calling unicast networkMonitor constructor" );

			// this is a static variable, but initialize() is called once for every
			// node. So we need to protect this, to prevent multiple instantiations.			
	        if(networkMonitor == null)
	           networkMonitor = new NetworkMonitor (port.getValue ());
    	}
    	else
    	{
			if (DEBUG) System.out.println 
 			("   EspduWriteTransform:  calling multicast networkMonitor constructor" );
			// this is a static variable, but initialize() is called once for every
			// node. So we need to protect this, to prevent multiple instantiations.			
	        if(networkMonitor == null)
	           networkMonitor = new NetworkMonitor (address.getValue (), 
 			                                        port.getValue ());
    	}
    	
		if (DEBUG) System.out.println
		     ("   EspduWriteTransform:  new Thread (networkMonitor)" );
		aThread = new Thread (networkMonitor);

		if (DEBUG) System.out.println 
 			      ("   EspduWriteTransform:  networkMonitor.setThread (aThread)" );
        networkMonitor.setThread (aThread);

        // from Ownship.java contructor:

        short entityID [] = {-1, -1, -1};

    // hi-fi vars
    pos     = new float[3]; // x, y, z
    posDt   = new float[3]; // aka: velocity
    ang     = new float[3]; // heading, pitch, roll
    angDt   = new float[3]; // aka: angular rates

    // lo-fi vars
    drPos   = new float[3]; // x, y, z
    drPosDt = new float[3]; // aka: velocity
    drAng   = new float[3]; // heading, pitch, roll
    drAngDt = new float[3]; // aka: angular rates

    // these are used for hpr->axis/angle conversions
    tmpQuat = new Quaternion();
    rot     = new float[4]; // axis/angle

    // now set init vars
    System.arraycopy(initPos, 0, pos, 0, 3);
    System.arraycopy(initPosDt, 0, posDt, 0, 3);
    System.arraycopy(initAng, 0, ang, 0, 3);
    System.arraycopy(initAngDt, 0, angDt, 0, 3);

    // now convert hpr (ang) to axis/angle (rot)
    tmpQuat.setEulers(ang);
    tmpQuat.getAxisAngle(rot);


		if (DEBUG) System.out.println ("   EspduWriteTransform:  aThread.start()" );
        aThread.start();  // should call run ()

        // When working properly we do return from thread start and this prints:
		if (DEBUG) System.out.println ("   EspduWriteTransform:  complete" );
  	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

	public void processEvents ( int count, Event events []) 
	{
		// ignore all but latest event so only one set of PDU network reads
		// 		and one set of dead-reckon calculations occur
		// this allows frequent updating triggered by the script eventIn

		if (DEBUG) System.out.println
		           ("processEvents:  " + count + " events queued");

		processEvent (events [count-1]);
	}
	
	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

	public void processEvent (Event event) 
	{
		EntityStatePdu		espdu	= new EntityStatePdu ();
		EntityID			entityID	= null;
		float				frameTime	= 0.0f;
		float				x, y, z, roll, pitch, yaw;

        SFVec3f 			position    = new SFVec3f ();

		orientation = new SFRotation ();
		eulers		= new float [3];
		rot 		= new float [4];

		if (DEBUG) System.out.println
				("  EspduWriteTransform:  processEvent");

		frameTime	= 1.0f / browser.getCurrentFrameRate ();
		dt 			= dt + frameTime;

		if (DEBUG) {
 		//	System.out.println ("  EspduWriteTransform:  browser.getCurrentSpeed () = "
 		//	                                           + browser.getCurrentSpeed ());
 			System.out.println ("  EspduWriteTransform:  browser.getCurrentFrameRate () = "
 			                                           + browser.getCurrentFrameRate ());
 			System.out.println ("  EspduWriteTransform:  cumulative dt = " + dt);
 		}

		pduCount ++;

		node = (Node) (transformNode.getValue()); // get VRML transformNode

 		if (DEBUG) System.out.println ("got transformNode.getValue() OK...");

		// verify proper espdu in hand??
        // ... awaiting multiple entity modifications by Don McGregor ...
		
 		if (DEBUG) System.out.println ("espdu name   = " + espdu.PDUName ());

// this fails, reason still unknown..
//		espdu.setMarking (((SFString) node.getExposedField ("marking")).getValue ());
//		if (DEBUG) System.out.println ("finished espdu.setMarking (..marking..) ");


		// Entity coordinate systems (right-hand rule applies)

		//          DIS                 VRML
		//
		//         -Z entity up         Y entity up
		//          |                   |
		//          |                   |
		//          |                   |
		//          |                   |
		//          +------- X          +-------- X    nose of entity body
		//         /                   /
		//        /                   /
		//       /                   /
		//      Y                   Z                   
		//         right-hand side
		//         of entity body
		//
		// Rotation angle axes (right-hand rule applies):
		//
		//          DIS                 VRML          Rotation angle synonyms
		//
		// Roll     X                    X            phi
		// Pitch    Y                    Z            theta
		// Yaw		Z					-Y            psi

		// Calculate VRML position/orientation from DIS position/orientation:
		
		position = (SFVec3f) node.getExposedField ("translation");	// get
		if (DEBUG) System.out.println("transformNode.translation = " + position);

        x = position.getX ();
        y = position.getY ();
        z = position.getZ ();

        espdu.setEntityLocationX (x);
        espdu.setEntityLocationY (y);
        espdu.setEntityLocationZ (z);

		if (DEBUG) System.out.println("position = (" + x + ", " + y + ", " + z + ")");

        espdu.setEntityLinearVelocityX (0.0f);
        espdu.setEntityLinearVelocityY (0.0f);
        espdu.setEntityLinearVelocityZ (0.0f);

		orientation = (SFRotation) node.getExposedField ("rotation");	// get
		if (DEBUG) System.out.println("transformNode.rotation = " + orientation);

		orientation.getValue (rot);

		if (DEBUG) System.out.println("rotation values = (" + rot[0] + ", " +
		                               rot[1] + ", " + rot[2] + ", " + rot[3] + ")" );

		quaternion.setAxisAngle (rot);
		
		if (DEBUG) { System.out.print ("quaternion " );  quaternion.print (); }

		quaternion.getEulers (eulers);

		// note Kent's quaternion code has irregular ordering of Euler angles
		// which (by whatever method :) accomplishes the angle transformation
		// desired...   (needs to be verified using NPS AUV)
				
		yaw   = -eulers [0];
		roll  =  eulers [1];
		pitch =  eulers [2];

		// set roll, pitch, and yaw in the PDU
		
        espdu.setEntityOrientationPsi (roll);
        espdu.setEntityOrientationTheta (pitch);
        espdu.setEntityOrientationPhi (yaw);

		if (DEBUG) System.out.println ("phi/theta/psi = "
				+ roll  * 180.0 / Math.PI + ", "
				+ pitch * 180.0 / Math.PI + ", "
				+ yaw   * 180.0 / Math.PI + " degrees");

		// This method is called at writeInterval durations
		// no dead reckoning needed, just use values in the VRML scene
		// we are here in the processEvent method so it is OK to send the PDU

		espdu.printValues (2, System.out);  // (indentLevel, printStream)

        networkMonitor.sendPDU (espdu, address.getValue(), port.getValue());
		
		return;
	}
}

