/*
 File:		DetonationPdu.java
 CVS Info:	$Id: DetonationPdu.java,v 1.2 1998/01/27 18:44:05 mcgredo Exp $
 Compiler:	jdk 1.1.6 
 */


package mil.navy.nps.dis;                                   // package to which we belong

import mil.navy.nps.util.*;         // General-purpose utilities
import mil.navy.nps.disEnumerations.*;         // Enumerations for DIS

import java.lang.*;
import java.util.*;                                         // utility stuff we need
import java.io.*;                                           // input/output for serialization

import iwt.*;
import org.w3c.dom.*;

/**
 *@version 1.0
 *@author <a href="mailto:mcgredo@stl.nps.navy.mil"> Don McGregor</a> (<a href="http://www.stl.nps.navy.mil/~mcgredo">http://www.stl.nps.navy.mil/~mcgredo</a>)
 *
 *<dt><b>Location:</b>
 *<dd><a href="http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/DetonationPdu.java">
 *  http://www.stl.nps.navy.mil/dis-java-vrml/mil/navy/nps/dis/DetonationPdu.java</a>
 *
 *<dt><b>Hierarchy Diagram:</b>
 *<dd><a href="images/PduClassHierarchy.jpg"><IMG SRC="images/PduClassHierarchyButton.jpg" ALIGN=ABSCENTER></a>
 *
 *<dt><b>Summary:</b>
 *<dd> The detonation or impact of munitions shall be communicated by issuing a Detonation PDU.
 *
 *<dt><b>Explanation:</b>
 *<dd>The detonation Pdu denotes the detonation of a weapon. It inherits the header
 *  information from ProtocolDataUnit, an abstract class that contains
 *  assorted protocol information. It implements the IDs of what's
 *  firing, what's being shot at, the munition, the event, and where
 *  the munition is headed.
 *
 *  As with other PDUs, it knows how to serialize and deserialize itself
 *  from the wire. It also knows how to clone itself, and knows how to
 *  calculate its size when sent to the wire.
 *
 *<dt><b>History:</b>
 *<dd>		16Dec96	/Don McGregor		/New
 *<dd>		10Mar97	/Don McGregor		/Cleaned up for javadoc
 *<dd>		16Apr97	/Don McGregor		/PrintStream passed to printValues
 *<dd>		8Dec97	/Ronan Fauglas		/changes for documentation templates + complements in documentation
 *<dd>		11Dec97	/Ronan Fauglas		/changed access methods: thisVariable() --> getThisVariable()
 *
 *<dt><b>References:</b>
 *<dd>		DIS Data Dictionary :<a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/84.htm">
 *		Detonation PDU</a>	
 *<dd>		DIS-Java-VRML Working Group: <a href="http://www.stl.nps.navy.mil/dis-java-vrml/"> 
 *		http://www.stl.nps.navy.mil/dis-java-vrml/</a>
 *<dd>		DIS specification : IEEE 1278.1, Section 4.4.3.3, 5.4.4.2
 *
 *@see ProtocolDataUnit
 *@see PduElement 
 *@see SerializationInterface
 *@see FirePdu
 */
public class DetonationPdu extends ProtocolDataUnit 
{
    /**
     *Firing Entity Identification - This field shall identify the firing entity.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>If the detonation is not preceded by a Fire PDU then the Firing Entity Identification shall be
     *NO SPECIFIC ENTITY.
     *</dl>
     */
    protected EntityID          firingEntityID;             // ID of entity that's doing the shooting

    /**
     *Target Entity Identification - This field shall identify the target entity. 
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd> If the target ID is unknown, this field shall contain the value TARGET ID UNKNOWN.
     *</dl>
     */
    protected EntityID          targetEntityID;             // ID of entity that is undergoing darwinian selection

    /**
     *Munition Identification - This field shall specify the entity ID of the fired munition 
     *if tracking data is required.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd> A munition ID shall have a value of MUNITION NOT TRACKED if tracking data for the munition
     *  is not required.
     *</dl>
     */
    protected EntityID          munitionID;                 // ID of munition being shot

    /**
     *Event Identification - This field shall contain the same data as in the Event Identification field 
     *of the Fire PDU that communicated the launch of the munition. 
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>If the detonation is not preceded by a corresponding fire event, then the Event Number field
     *of the Event Identifier record shall be zero (e.g., land mines detonation).
     *</dl>
     */
    protected EventID           eventID;                    // ID of event

    /**
     *Velocity - This field shall specify the velocity of the munition immediately before detonation/impact. 
     */
    protected LinearVelocity          velocity;                   // Velocity of muntion just before detonation/impact
    
    /**
     *Location in World Coordinate - This field shall specify the location of the detonation 
     *in world coordinates.
     */
    protected WorldCoordinate     locationInWorldCoordinate;                   // x,y,z location, in 64 bit coords
    
    /**
     *Burst Descriptor - This field shall describe the type of munition impacting or detonating, 
     *the warhead, the fuse, the quantity, and the rate.
     */
    protected BurstDescriptor   burstDescription;           // info about burst

    /**
     *Location in Entity's coordinates - This field shall specify the location of the detonation 
     *or impact in the target entity's coordinate system. This information should
     *be used for damage assessment.
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>  If the ID of the target is unknown, this field shall contain NO LOCATION.
     *</dl>
     */
    protected EntityCoordinate    locationInEntityCoordinates;             // location of impact, in entity local coords

    /**
     *Detonation Result - This field shall specify the result of the detonation. 
     *
     *<dl>
     *<dt><b>Value:</b>
     *<dd>  Enumeration. See information below for more information
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary: <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/21.htm">
     *		Force ID Field</a>
     *<dd>	See Section 5 in EBV-DOC
     *</dl>
     */
    protected UnsignedByte      detonationResult;           // Blood & guts, usually


    //protected UnsignedByte        articulationParamterCount;  // not used; we query the list for this instead
    //protected UnsignedShort       padding;                // unused; we write out padding directly in serialize

    /**
     *List of articulation parameters.
     *
     *<dl>
     *<dt><b>References:</b>
     *<dd>	DIS Data Dictionary: <a href="http://stds.sc.ist.ucf.edu/dis/dis-dd/pdu/6d.htm">
     *		Articulation Parameter Record</a>
     *<dd>	See Section 5 in EBV-DOC
     *</dl>
     */
    protected Vector            articulationParameters;     // List of articulation parameters

    /**
     *Constant value--size of an Detonation PDU with header, without the articulation parameters.
     *<code>sizeOf = 104 bytes</code>
     */
    public static final int   sizeOf = 104;                 // size of "stripped" DetonationPDU, AS WRITTEN TO WIRE


/**
 *Default constructor--fills with zeros for all values.
 */
public DetonationPdu()
{
    super.setPduType(PDUTypeField.DETONATION);

    firingEntityID  = new EntityID();
    targetEntityID  = new EntityID();
    munitionID      = new EntityID();
    eventID         = new EventID();
    velocity        = new LinearVelocity(0,0,0);                  // special nerf ammo by default--hits at zero speed.
                                                            // let's play nice!

    locationInWorldCoordinate        = new WorldCoordinate(0, 0, 0);

    burstDescription = new BurstDescriptor();

    locationInEntityCoordinates = new EntityCoordinate(0,0,0);
    detonationResult = new UnsignedByte(0);
    articulationParameters = new Vector();

    return;
}

public Object clone()
{
    /** make a copy of the object. This requires a deep copy, so we don't have two
        objects sharing pointers to the same data.
    */

 DetonationPdu  newDetonationPdu = (DetonationPdu)super.clone();
 int            idx = 0;

 newDetonationPdu.setFiringEntityID(this.getFiringEntityID());
 newDetonationPdu.setTargetEntityID(this.getTargetEntityID());
 newDetonationPdu.setMunitionID(this.getMunitionID());
 newDetonationPdu.setEventID(this.getEventID());
 newDetonationPdu.setVelocity(this.getVelocity());

 newDetonationPdu.setLocationInWorldCoordinate(this.getLocationInWorldCoordinate());

 newDetonationPdu.setBurstDescription(this.getBurstDescription());
 newDetonationPdu.setDetonationResult(this.getDetonationResult());
 
 // Each of the articulation parameters needs to be cloned as well--it's not
 // good enough to just clone the Vector alone.

 for(idx = 0; idx < articulationParameters.size(); idx++)
 {
    newDetonationPdu.addArticulationParameter(this.getArticulationParameterAt(idx));
 }

 return newDetonationPdu;
}

public void serialize(DataOutputStream outputStream)
{

    UnsignedShort   padding = new UnsignedShort(0);
    UnsignedByte    parameterCount;
    int             idx;

    // write out the data to an output stream. Order is important here, since
    // it needs to conform to the DIS standard.

    super.serialize(outputStream);      // write out header info
    

    firingEntityID.serialize(outputStream);
    targetEntityID.serialize(outputStream);
    munitionID.serialize(outputStream);
    eventID.serialize(outputStream);
    velocity.serialize(outputStream);
    locationInWorldCoordinate.serialize(outputStream);
    burstDescription.serialize(outputStream);
    locationInEntityCoordinates.serialize(outputStream);
    detonationResult.serialize(outputStream);

    // We synthisize these from the list data or out of thin air
    parameterCount = new UnsignedByte(articulationParameters.size());

    parameterCount.serialize(outputStream);
    padding.serialize(outputStream);                // aligns us w/ byte boundary

    for(idx = 0; idx < articulationParameters.size(); idx++)
    {
        ArticulationParameter  aParameter;

        aParameter = (ArticulationParameter)articulationParameters.elementAt(idx);
        aParameter.serialize(outputStream);
    }

        
   return;
}

public void  deSerialize(DataInputStream inputStream)
{
 /**
    deserialize our data from the stream. We first call the superclass,
    so that all the data there is read into the buffer first.
    Then we deserialize each of our ivars to the stream, being careful
    about the order.
   */

    int             idx = 0;
    UnsignedByte    articulationParameterCount = new UnsignedByte();
    UnsignedShort   padding = new UnsignedShort(0);

    super.deSerialize(inputStream);

    // Do our ivars

    
    firingEntityID.deSerialize(inputStream);
    targetEntityID.deSerialize(inputStream);
    munitionID.deSerialize(inputStream);
    eventID.deSerialize(inputStream);
    velocity.deSerialize(inputStream);
    locationInWorldCoordinate.deSerialize(inputStream);
    burstDescription.deSerialize(inputStream);
    locationInEntityCoordinates.deSerialize(inputStream);
    detonationResult.deSerialize(inputStream);

    articulationParameterCount.deSerialize(inputStream);
    padding.deSerialize(inputStream);

    for(idx = 0; idx < articulationParameterCount.intValue(); idx++)
    {
        ArticulationParameter  aParameter = new ArticulationParameter();

        aParameter.deSerialize(inputStream);
        articulationParameters.addElement(aParameter);
            
    }

    return;
}

public int length()
{
    // basic size + size of list of parameters.

    return sizeOf + (articulationParameters.size() * ArticulationParameter.sizeOf);
}

public String PDUName()
{
  return new String("Detonation PDU");
}

public void printValues(int indentLevel, PrintStream printStream)
{
    // print the values of the object out, with correct level of
    // indentation on the page.

    char[]  indent = new char[indentLevel];
    int     idx = 0;
    String  spacing;
    int     superclassIndent = indentLevel;

    for(idx = 0; idx < indentLevel; idx++)
        indent[idx] = ' ';

    spacing = new String(indent);

    printStream.println();
    printStream.println("Detonation PDU-");

    if(superclassIndent > 0)
      superclassIndent -= 1;

    super.printValues(superclassIndent, printStream);

    firingEntityID.printValues(indentLevel, printStream);
    targetEntityID.printValues(indentLevel, printStream);
    munitionID.printValues(indentLevel, printStream);
    eventID.printValues(indentLevel, printStream);
    velocity.printValues(indentLevel, printStream);
    locationInWorldCoordinate.printValues(indentLevel, printStream);
    burstDescription.printValues(indentLevel, printStream);
    locationInEntityCoordinates.printValues(indentLevel, printStream);
    printStream.println(indent + "detonationResult: "    + detonationResult.intValue());
    printStream.println(indent + "Parameter count: " + articulationParameters.size());

    for(idx = 0; idx < articulationParameters.size(); idx ++)
     {
        ArticulationParameter   aParam;

        aParam = (ArticulationParameter)articulationParameters.elementAt(idx);
        aParam.printValues(indentLevel + 4, printStream);                // indent the parts of us a bit
     }

    return;
}

    //Accessor methods

public EntityID getFiringEntityID()
{   return (EntityID)firingEntityID.clone();
}
public void setFiringEntityID(EntityID pFiringEntityID)
{firingEntityID = pFiringEntityID;
}

public EntityID getTargetEntityID()
{   return (EntityID)targetEntityID.clone();
}
public void setTargetEntityID(EntityID pTargetEntityID)
{targetEntityID = pTargetEntityID;
}

public EntityID getMunitionID()
{   return (EntityID)munitionID.clone();
}
public void setMunitionID(EntityID pMunitionID)
{munitionID = pMunitionID;
}

public EventID getEventID()
{   return (EventID)eventID.clone();
}
public void setEventID(EventID pEventID)
{eventID = pEventID;
}


public LinearVelocity getVelocity()
{   return (LinearVelocity)velocity.clone();
}
public void setVelocity(LinearVelocity pVelocity)
{velocity = pVelocity;
}
public void setVelocity(float x, float y, float z)
{
    velocity = new LinearVelocity(x, y, z);
}

public WorldCoordinate getLocationInWorldCoordinate()
{   return (WorldCoordinate)locationInWorldCoordinate.clone();
}
public void setLocationInWorldCoordinate(WorldCoordinate pLocation)
{locationInWorldCoordinate = pLocation;
}
public void setLocationInWorldCoordinate(double x, double y, double z)
{
    locationInWorldCoordinate = new WorldCoordinate(x, y, z);
}

public BurstDescriptor getBurstDescription()
{   return (BurstDescriptor)burstDescription.clone();
}
public void setBurstDescription(BurstDescriptor pBurstDescription)
{burstDescription = pBurstDescription;
}

public EntityCoordinate getLocationInEntityCoordinates()
{   return (EntityCoordinate)locationInEntityCoordinates.clone();
}
public void setLocationInEntityCoordinates(EntityCoordinate pLocationInEntityCoordinates)
{locationInEntityCoordinates = pLocationInEntityCoordinates;
}
public void setLocationInEntityCoordinates(float x, float y, float z)
{
    locationInEntityCoordinates = new EntityCoordinate(x, y, z);
}

public UnsignedByte getDetonationResult()
{   return (UnsignedByte)detonationResult.clone();
}
public void setDetonationResult(UnsignedByte pDetonationResult)
{detonationResult = pDetonationResult;
}
public void setDetonationResult(int pDetonationResult)
{ detonationResult = new UnsignedByte(pDetonationResult);
}

// Tam Pulikal
public UnsignedByte articulationParameterCount()
{ return new UnsignedByte(articulationParameters.size());
}

/**
 *Returns the Articulation Parameter at the index given in parameter.
 *
 *@param pIdx the index of the Articulation Parameter we want to obtain
 *@return  the Articulation Parameter pointed by the index in parameter
 */
public ArticulationParameter getArticulationParameterAt(int pIdx)
{
 ArticulationParameter  aParameter;

  // return a copy of the pIdx-th articulation parameter.

 aParameter = (ArticulationParameter)articulationParameters.elementAt(pIdx);

 return (ArticulationParameter)aParameter.clone();
}
 
/**
 *Inserts a Articulation Parameter at the end of the list.
 *
 *@param pParameter the Articulation Parameter to be inserted in the list
 */
public void addArticulationParameter(ArticulationParameter pParameter)
{
 // Add one articulation parameter to the end, and increment the parameter
 // count accordingly.

 articulationParameters.addElement(pParameter);
}

  public void
  write(org.w3c.dom.Document outDoc)
  {
     insert(outDoc);

  }// end of write()

  public void
  read(org.w3c.dom.Document inDoc) throws InvalidDocumentType
  {
  }// end of read()

  public void
  insert(org.w3c.dom.Node outNode)
  {

    UnsignedShort   padding = new UnsignedShort(0);
    UnsignedByte    parameterCount;
    int             idx;

    Element         ele;
    Element         ele1;
    Text            txt1;

    //System.out.println("Entering Fire serialization");
    Document outDoc = outNode.getOwnerDocument();
    ele1 = outDoc.createElement(iwt.DIS_TAGS.PDU);
    outNode.appendChild(ele1);

    // write out the data to an output stream. Order is important here, since
    // it needs to conform to the DIS standard.

    // write out header info
    insertHeader(ele1);
    
    ele = outDoc.createElement(DIS_TAGS.FiringEntity);
    ele1.appendChild(ele);
    firingEntityID.insert(ele);

    ele = outDoc.createElement(DIS_TAGS.TargetEntity);
    ele1.appendChild(ele);
    targetEntityID.insert(ele);

    ele = outDoc.createElement(DIS_TAGS.Munition);
    ele1.appendChild(ele);
    munitionID.insert(ele);

    eventID.insert(ele1);

    velocity.insert(ele1);

    locationInWorldCoordinate.insert(ele1);

    burstDescription.insert(ele1);

    locationInEntityCoordinates.insert(ele1);

    ele = outDoc.createElement(DIS_TAGS.DetonationResult);
    ele.setAttribute(DIS_TAGS.Value,detonationResult.toString());
    ele1.appendChild(ele);

    // We synthisize these from the list data or out of thin air
    parameterCount = new UnsignedByte(articulationParameters.size());

    ele = outDoc.createElement(DIS_TAGS.ArticulationParameters);
    ele.setAttribute(DIS_TAGS.Value,parameterCount.toString());
    ele1.appendChild(ele);

    for(idx = 0; idx < articulationParameters.size(); idx++)
    {
        ArticulationParameter  aParameter;

        aParameter = (ArticulationParameter)articulationParameters.elementAt(idx);
        aParameter.insert(ele);
    }

  }// end of insert()

  public void
  extract(org.w3c.dom.Node inNode) throws InvalidDocumentType
  {
     Element         ele;
     NodeList        nodes;
     NodeList        aList;
     int i=0;

     if( !inNode.getNodeName().equals(DIS_TAGS.PDU) ) 
         throw new InvalidDocumentType();

     nodes = inNode.getChildNodes();

     while( i<nodes.getLength() )
     {
        Node aNode = nodes.item(i);

        if( aNode!=null )
        {

            if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                aNode.getNodeName().equals(iwt.DIS_TAGS.DIS_PDU_Header) )
                extractHeader(aNode);

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.DetonationResult) )
                 {
                     String val1 = ((Element)aNode).getAttribute(DIS_TAGS.Value)
;
                     try
                     {  detonationResult = new UnsignedByte(Integer.parseInt(val1
));
                     } catch(NumberFormatException nfe) {}
                 }

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.EventID) )
                     eventID.extract(aNode);

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.EntityCoordinate) )
                     locationInEntityCoordinates.extract(aNode);

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.WorldCoordinate) )
                     locationInWorldCoordinate.extract(aNode);

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.BurstDescriptor) )
                     burstDescription.extract(aNode);

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.LinearVelocity) )
                     velocity.extract(aNode);

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.Munition) )

                 {/*
                     aList = ((Element)aNode).getElementsByTagName(DIS_TAGS.EntityID);

                     if( aList!=null && aList.getLength()>0 )
                     {
                        Node bNode = nodes.item(0);
                        if( bNode!=null &&
                            bNode.getNodeType()==org.w3c.dom.Node.ELEMENT_NODE)
                            munitionID.extract(bNode);

                     }// end of if
					 */
					 int k = 0;
					 aList = aNode.getChildNodes();
					 if( aList!=null )
						while( k<aList.getLength() ) 
						{
							Node bNode = aList.item(k);
							if( bNode!=null && bNode.getNodeName().equals(iwt.DIS_TAGS.EntityID) ) 
							{ 
							munitionID.extract(bNode);
							}// end of if 
							k++;
						}// end of while

                 }// end of if

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.FiringEntity))

                 {/*

                     aList = ((Element)aNode).getElementsByTagName(DIS_TAGS.EntityID);

                     if( aList!=null && aList.getLength()>0 )
                     {
                        Node bNode = nodes.item(0);
                        if( bNode!=null &&
                            bNode.getNodeType()==org.w3c.dom.Node.ELEMENT_NODE)
                            firingEntityID.extract(bNode);

                     }// end of if */
					 int k = 0;
					 aList = aNode.getChildNodes();
					 if( aList!=null )
						while( k<aList.getLength() ) 
						{
							Node bNode = aList.item(k);
							if( bNode!=null && bNode.getNodeName().equals(iwt.DIS_TAGS.EntityID) ) 
							{ 
							firingEntityID.extract(bNode);
							}// end of if 
							k++;
						}// end of while

                 }// end of if

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.TargetEntity))

                 {/*
                     aList = ((Element)aNode).getElementsByTagName(DIS_TAGS.EntityID);

                     if( aList!=null && aList.getLength()>0 )
                     {
                        Node bNode = nodes.item(0);
                        if( bNode!=null &&
                            bNode.getNodeType()==org.w3c.dom.Node.ELEMENT_NODE)
                            targetEntityID.extract(bNode);

                     }// end of if */
					 int k = 0;
					 aList = aNode.getChildNodes();
					 if( aList!=null )
						while( k<aList.getLength() ) 
						{
							Node bNode = aList.item(k);
							if( bNode!=null && bNode.getNodeName().equals(iwt.DIS_TAGS.EntityID) ) 
							{ 
							targetEntityID.extract(bNode);
							}// end of if 
							k++;
						}// end of while


                 }// end of if

            else if( aNode.getNodeType() == Node.ELEMENT_NODE &&
                     aNode.getNodeName().equals(iwt.DIS_TAGS.ArticulationParameters))

                 {
                     String val1 = ((Element)aNode).getAttribute(iwt.DIS_TAGS.Value);
                     int k=0, lim = 0;

                     try
                     {  lim = Integer.parseInt(val1);
                     } catch( NumberFormatException nfe) {}

                     aList = ((Element)aNode).getElementsByTagName(DIS_TAGS.ArticulationParameter);

                     articulationParameters = new Vector();

                     if( aList!=null )
                     for(k=0;k<lim&&k<aList.getLength();k++) 
                     {
                        Node bNode = nodes.item(k);
                        if( bNode!=null &&
                            bNode.getNodeType() == Node.ELEMENT_NODE &&
                            bNode.getNodeName().equals(iwt.DIS_TAGS.ArticulationParameter) )
                        {
                          ArticulationParameter aPa = new ArticulationParameter();
                          aPa.extract(bNode);
                          articulationParameters.addElement(aPa);

                        }// end of if
              
                     }// end of for

                 }// end of if

        }// end of if

        i++;

     }// end of while

  }// end of read()

} // end of class DetonationPdu

